# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.binning import (
    UsesBinning,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.pdf import (
    EnergyPDF,
    PDFAxis,
)
from skyllh.core.py import (
    classname,
)
from skyllh.core.smoothing import (
    UNSMOOTH_AXIS,
    SmoothingFilter,
    HistSmoothingMethod,
    NoHistSmoothingMethod,
    NeighboringBinHistSmoothingMethod,
)
from skyllh.core.timing import (
    TaskTimer,
)

logger = get_logger(__name__)


class I3EnergyPDF(
        EnergyPDF,
        UsesBinning,
):
    """This is the base class for all IceCube specific energy PDF models.
    IceCube energy PDFs depend soley on the energy and the
    zenith angle, and hence, on the declination of the event.

    The IceCube energy PDF is modeled as a 1d histogram in energy,
    but for different sin(declination) bins, hence, stored as a 2d histogram.
    """
    def __init__(
            self,
            pmm,
            data_log10_energy,
            data_sin_dec,
            data_mcweight,
            data_physicsweight,
            log10_energy_binning,
            sin_dec_binning,
            smoothing_filter,
            **kwargs,
    ):
        """Creates a new IceCube energy PDF object.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper | None
            The instance of ParameterModelMapper defining the global parameters
            and their mapping to local model/source parameters.
            It can be ``None``, if the PDF does not depend on any parameters.
        data_log10_energy : 1d ndarray
            The array holding the log10(E) values of the events.
        data_sin_dec : 1d ndarray
            The array holding the sin(dec) values of the events.
        data_mcweight : 1d ndarray
            The array holding the monte-carlo weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        data_physicsweight : 1d ndarray
            The array holding the physics weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        log10_energy_binning : instance of BinningDefinition
            The binning definition for the log10(E) axis.
        sin_dec_binning : instance of BinningDefinition
            The binning definition for the sin(declination) axis.
        smoothing_filter : instance of SmoothingFilter | None
            The smoothing filter to use for smoothing the energy histogram.
            If ``None``, no smoothing will be applied.
        """
        super().__init__(
            pmm=pmm,
            **kwargs)

        # Define the PDF axes.
        self.add_axis(
            PDFAxis(
                name='log_energy',
                vmin=log10_energy_binning.lower_edge,
                vmax=log10_energy_binning.upper_edge))
        self.add_axis(
            PDFAxis(
                name='sin_dec',
                vmin=sin_dec_binning.lower_edge,
                vmax=sin_dec_binning.upper_edge))

        self.add_binning(log10_energy_binning, 'log_energy')
        self.add_binning(sin_dec_binning, 'sin_dec')

        # Create the smoothing method instance tailored to the energy PDF.
        # We will smooth only the first axis (log10(E)).
        if (smoothing_filter is not None) and\
           (not isinstance(smoothing_filter, SmoothingFilter)):
            raise TypeError(
                'The smoothing_filter argument must be None or an instance of '
                f'SmoothingFilter! It is of type {classname(smoothing_filter)}')
        if smoothing_filter is None:
            self.hist_smoothing_method = NoHistSmoothingMethod()
        else:
            self.hist_smoothing_method = NeighboringBinHistSmoothingMethod(
                (smoothing_filter.axis_kernel_array, UNSMOOTH_AXIS))

        # We have to figure out, which histogram bins are zero due to no
        # monte-carlo coverage, and which due to zero physics model
        # contribution.

        # Create a 2D histogram with only the MC events to determine the MC
        # coverage.
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_log10_energy,
            data_sin_dec,
            bins=[log10_energy_binning.binedges, sin_dec_binning.binedges],
            range=[log10_energy_binning.range, sin_dec_binning.range],
            density=False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered = h > 0

        # Select the events which have MC coverage but zero physics
        # contribution, i.e. the physics model predicts zero contribution.
        mask = data_physicsweight == 0.

        # Create a 2D histogram with only the MC events that have zero physics
        # contribution. Note: By construction the zero physics contribution bins
        # are a subset of the MC covered bins.
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_log10_energy[mask],
            data_sin_dec[mask],
            bins=[log10_energy_binning.binedges, sin_dec_binning.binedges],
            range=[log10_energy_binning.range, sin_dec_binning.range],
            density=False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered_zero_physics = h > 0

        # Create a 2D histogram with only the data which has physics
        # contribution. We will do the normalization along the logE
        # axis manually.
        data_weights = data_mcweight[~mask] * data_physicsweight[~mask]
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_log10_energy[~mask],
            data_sin_dec[~mask],
            bins=[log10_energy_binning.binedges, sin_dec_binning.binedges],
            weights=data_weights,
            range=[log10_energy_binning.range, sin_dec_binning.range],
            density=False)

        # Calculate the normalization for each logE bin. Hence we need to sum
        # over the logE bins (axis 0) for each sin(dec) bin and need to divide
        # by the logE bin widths along the sin(dec) bins. The result array norm
        # is a 2D array of the same shape as h.
        norms = np.sum(h, axis=(0,))[np.newaxis, ...] *\
            np.diff(log10_energy_binning.binedges)[..., np.newaxis]
        h /= norms
        h = self._hist_smoothing_method.smooth(h)

        self._hist_log10_energy_sin_dec = h

    @property
    def hist_smoothing_method(self):
        """The HistSmoothingMethod instance defining the smoothing filter of the
        energy PDF histogram.
        """
        return self._hist_smoothing_method

    @hist_smoothing_method.setter
    def hist_smoothing_method(self, method):
        if not isinstance(method, HistSmoothingMethod):
            raise TypeError(
                'The hist_smoothing_method property must be an instance of '
                f'HistSmoothingMethod! It is of type {classname(method)}')
        self._hist_smoothing_method = method

    @property
    def hist(self):
        """(read-only) The 2D logE-sinDec histogram array.
        """
        return self._hist_log10_energy_sin_dec

    @property
    def hist_mask_mc_covered(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage.
        """
        return self._hist_mask_mc_covered

    @property
    def hist_mask_mc_covered_zero_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage but zero physics
        contribution.
        """
        return self._hist_mask_mc_covered_zero_physics

    @property
    def hist_mask_mc_covered_with_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage and has physics
        contribution.
        """
        mask = (
            self._hist_mask_mc_covered & ~self._hist_mask_mc_covered_zero_physics
        )
        return mask

    def assert_is_valid_for_trial_data(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Checks if this energy PDF is valid for all the given trial events.
        It checks if all the data is within the log10(E) and sin(dec) binning
        range.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events.
            The following data fields must exist:

            log_energy : float
                The base-10 logarithm of the energy value of the data event.
            dec : float
                The declination of the data event.

        tl : instance of TimeLord | None
            The optional instance of TimeLord for measuring timing information.

        Raises
        ------
        ValueError
            If some of the data is outside the log10(E) or sin(dec) binning
            range.
        """
        log10_energy_binning = self.get_binning('log_energy')
        sin_dec_binning = self.get_binning('sin_dec')

        data_log10_energy = tdm['log_energy']
        data_sin_dec = np.sin(tdm['dec'])

        if log10_energy_binning.any_data_out_of_range(data_log10_energy):
            oor_data = log10_energy_binning.get_out_of_range_data(
                data_log10_energy)
            raise ValueError(
                'Some data is outside the log10(E) range '
                f'({log10_energy_binning.lower_edge:.3f},'
                f' {log10_energy_binning.upper_edge:.3f})! '
                f'The following data values are out of range: {oor_data}')

        if sin_dec_binning.any_data_out_of_range(data_sin_dec):
            oor_data = sin_dec_binning.get_out_of_range_data(
                data_sin_dec)
            raise ValueError(
                'Some data is outside the sin(dec) range '
                f'({sin_dec_binning.lower_edge:.3f},'
                f' {sin_dec_binning.upper_edge:.3f})! '
                f'The following data values are out of range: {oor_data}')

    def get_pd(self, tdm, params_recarray=None, tl=None):
        """Calculates the energy probability density of each event.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the data events for which the
            probability density should be calculated.
            The following data fields must exist:

            log_energy : float
                The base-10 logarithm of the energy value of the event.
            sin_dec : float
                The sin(declination) value of the event.

        params_recarray : None
            Unused interface parameter.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of ndarray
            The 1D (N_events,)-shaped numpy ndarray with the energy probability
            density for each event.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each fit parameter. The key of the dictionary is the id
            of the global fit parameter. Because this energy PDF does not depend
            on any fit parameters, an empty dictionary is returned.
        """
        log10_energy_binning = self.get_binning('log_energy')
        sin_dec_binning = self.get_binning('sin_dec')

        log10_energy_idx = np.digitize(
            tdm['log_energy'], log10_energy_binning.binedges) - 1
        sin_dec_idx = np.digitize(
            tdm['sin_dec'], sin_dec_binning.binedges) - 1

        with TaskTimer(tl, 'Evaluating log10_energy-sin_dec histogram.'):
            pd = self._hist_log10_energy_sin_dec[
                (log10_energy_idx, sin_dec_idx)]

        return (pd, dict())



# -*- coding: utf-8 -*-

"""The :mod:`~skyllh.core.model` module contains the base class ``SourceModel``
for modelling a source in the sky. What kind of properties this source has is
modeled by a derived class. The most common one is the PointLikeSource source
model for a point-like source at a given location in the sky.
"""

import numpy as np

from skyllh.core.model import (
    Model,
    ModelCollection,
)
from skyllh.core.py import (
    classname,
    float_cast,
    issequenceof,
    str_cast,
    typename,
)


class SourceModel(
        Model,
):
    """The base class for all source models in SkyLLH. A source can have a
    relative weight w.r.t. other sources.
    """
    def __init__(
            self,
            name=None,
            classification=None,
            weight=None,
            **kwargs):
        """Creates a new source model instance.

        Parameters
        ----------
        name : str | None
            The name of the source model.
        classification : str | None
            The astronomical classification of the source.
        weight : float | None
            The relative weight of the source w.r.t. other sources.
            If set to None, unity will be used.
        """
        super().__init__(
            name=name,
            **kwargs)

        self.classification = classification
        self.weight = weight

    @property
    def classification(self):
        """The astronomical classification of the source.
        """
        return self._classification

    @classification.setter
    def classification(self, c):
        self._classification = str_cast(
            c,
            'The classification property must be castable to type str!',
            allow_None=True)

    @property
    def weight(self):
        """The weight of the source.
        """
        return self._weight

    @weight.setter
    def weight(self, w):
        if w is None:
            w = 1.
        w = float_cast(
            w,
            'The weight property must be castable to type float!')
        self._weight = w


class SourceModelCollection(
        ModelCollection,
):
    """This class describes a collection of source models. It can be used to
    group sources into a single object, for instance for a stacking analysis.
    """
    @staticmethod
    def cast(
            obj,
            errmsg=None,
            **kwargs):
        """Casts the given object to a SourceModelCollection object. If the cast
        fails, a TypeError with the given error message is raised.

        Parameters
        ----------
        obj : SourceModel | sequence of SourceModel | SourceModelCollection |
                None
            The object that should be casted to SourceModelCollection.
            If set to None, an empty SourceModelCollection is created.
        errmsg : str | None
            The error message if the cast fails.
            If set to None, a generic error message will be used.

        Additional keyword arguments
        ----------------------------
        Additional keyword arguments are passed to the constructor of the
        SourceModelCollection class.

        Raises
        ------
        TypeError
            If the cast failed.
        """
        if obj is None:
            return SourceModelCollection(
                sources=None, source_type=SourceModel, **kwargs)

        if isinstance(obj, SourceModel):
            return SourceModelCollection(
                sources=[obj], source_type=SourceModel, **kwargs)

        if isinstance(obj, SourceModelCollection):
            return obj

        if issequenceof(obj, SourceModel):
            return SourceModelCollection(
                sources=obj, source_type=SourceModel, **kwargs)

        if errmsg is None:
            errmsg = (f'Cast of object "{str(obj)}" of type '
                      f'"{typename(obj)}" to SourceModelCollection failed!')
        raise TypeError(errmsg)

    def __init__(
            self,
            sources=None,
            source_type=None,
            **kwargs):
        """Creates a new source collection.

        Parameters
        ----------
        sources : sequence of source_type instances | None
            The sequence of sources this collection should be initalized with.
            If set to None, an empty SourceModelCollection instance is created.
        source_type : type | None
            The type of the source.
            If set to None (default), SourceModel will be used.
        """
        if source_type is None:
            source_type = SourceModel

        super().__init__(
            models=sources,
            model_type=source_type,
            **kwargs)

    @property
    def source_type(self):
        """(read-only) The type of the source model.
        This property is an alias for the `obj_type` property.
        """
        return self.model_type

    @property
    def sources(self):
        """(read-only) The list of sources of type ``source_type``.
        """
        return self.models


class IsPointlike(
        object):
    """This is a classifier class that can be used by other classes to indicate
    that the specific class describes a point-like object.
    """
    def __init__(
            self,
            ra_func_instance=None,
            get_ra_func=None,
            set_ra_func=None,
            dec_func_instance=None,
            get_dec_func=None,
            set_dec_func=None,
            **kwargs):
        """Constructor method. Gets called when the an instance of a class is
        created which derives from this IsPointlike class.

        Parameters
        ----------
        ra_func_instance : object
            The instance object the right-ascention property's getter and setter
            functions are defined in.
        get_ra_func : callable
            The callable object of the getter function of the right-ascention
            property. It must have the call signature
            `__call__(ra_func_instance)`.
        set_ra_func : callable
            The callable object of the setter function of the right-ascention
            property. It must have the call signature
            `__call__(ra_func_instance, value)`.
        dec_func_instance : object
            The instance object the declination property's getter and setter
            functions are defined in.
        get_dec_func : object
            The callable object of the getter function of the declination
            property. It must have the call signature
            `__call__(dec_func_instance)`.
        set_dec_func : object
            The callable object of the setter function of the declination
            property. It must have the call signature
            `__call__(dec_func_instance, value)`.
        """
        super().__init__(**kwargs)

        self._ra_func_instance = ra_func_instance
        self._get_ra_func = get_ra_func
        self._set_ra_func = set_ra_func

        self._dec_func_instance = dec_func_instance
        self._get_dec_func = get_dec_func
        self._set_dec_func = set_dec_func

    @property
    def ra(self):
        """The right-ascention coordinate of the point-like source.
        """
        return self._get_ra_func(self._ra_func_instance)

    @ra.setter
    def ra(self, v):
        v = float_cast(
            v,
            'The ra property must be castable to type float!')
        self._set_ra_func(self._ra_func_instance, v)

    @property
    def dec(self):
        """The declination coordinate of the point-like source.
        """
        return self._get_dec_func(self._dec_func_instance)

    @dec.setter
    def dec(self, v):
        v = float_cast(
            v,
            'The dec property must be castable to type float!')
        self._set_dec_func(self._dec_func_instance, v)


class PointLikeSource(
        SourceModel,
        IsPointlike):
    """The PointLikeSource class is a source model for a point-like source
    object in the sky at a given location (right-ascention and declination).
    """
    def __init__(
            self,
            ra,
            dec,
            name=None,
            weight=None,
            **kwargs):
        """Creates a new PointLikeSource instance for defining a point-like
        source.

        Parameters
        ----------
        ra : float
            The right-ascention coordinate of the source in radians.
        dec : float
            The declination coordinate of the source in radians.
        name : str | None
            The name of the source.
        weight : float | None
            The relative weight of the source w.r.t. other sources.
            If set to None, unity will be used.
        """
        super().__init__(
            name=name,
            weight=weight,
            ra_func_instance=self,
            get_ra_func=type(self)._get_ra,
            set_ra_func=type(self)._set_ra,
            dec_func_instance=self,
            get_dec_func=type(self)._get_dec,
            set_dec_func=type(self)._set_dec,
            **kwargs,
        )

        self.ra = ra
        self.dec = dec

    def _get_ra(self):
        return self._ra

    def _set_ra(self, ra):
        self._ra = ra

    def _get_dec(self):
        return self._dec

    def _set_dec(self, dec):
        self._dec = dec

    def __str__(self):
        """Pretty string representation.
        """
        c = ''
        if self.classification is not None:
            c = f', classification={self.classification}'

        s = (
            f'{classname(self)}: "{self.name}": '
            '{ '
            f'ra={np.rad2deg(self.ra):.3f} deg, '
            f'dec={np.rad2deg(self.dec):.3f} deg'
            f'{c}'
            ' }'
        )

        return s



# -*- coding: utf-8 -*-

"""This module contains utility functions related analysis trials.
"""

import pickle

from skyllh.core.timing import (
    TaskTimer,
)


def create_pseudo_data_file(
        ana,
        rss,
        filename,
        mean_n_bkg_list=None,
        mean_n_sig=0,
        bkg_kwargs=None,
        sig_kwargs=None,
        tl=None
):
    """Creates a pickle file that contains the pseudo data for a single trial
    by generating background and signal events.

    Parameters
    ----------
    ana : Analysis
        The Analysis instance that should be used to generate the pseudo data.
    rss : RandomStateService
        The RandomStateService instance to use for generating random numbers.
    filename : str
        The data file name into which the generated pseudo data should get
        written to.
    mean_n_bkg_list : list of float | None
        The mean number of background events that should be generated for
        each dataset. If set to None (the default), the background
        generation method needs to obtain this number itself.
    mean_n_sig : float
        The mean number of signal events that should be generated for the
        trial. The actual number of generated events will be drawn from a
        Poisson distribution with this given signal mean as mean.
    bkg_kwargs : dict | None
        Additional keyword arguments for the `generate_events` method of the
        background generation method class. An usual keyword argument is
        `poisson`.
    sig_kwargs : dict | None
        Additional keyword arguments for the `generate_signal_events` method
        of the `SignalGenerator` class. An usual keyword argument is
        `poisson`.
    tl : TimeLord | None
        The instance of TimeLord that should be used to time individual tasks.

    """
    (n_bkg_events_list, bkg_events_list) = ana.generate_background_events(
        rss=rss,
        mean_n_bkg_list=mean_n_bkg_list,
        bkg_kwargs=bkg_kwargs,
        tl=tl
    )

    (n_sig, n_sig_events_list, sig_events_list) = ana.generate_signal_events(
        rss=rss,
        mean_n_sig=mean_n_sig,
        sig_kwargs=sig_kwargs,
        tl=tl
    )

    trial_data = dict(
        mean_n_bkg_list=mean_n_bkg_list,
        mean_n_sig=mean_n_sig,
        bkg_kwargs=bkg_kwargs,
        sig_kwargs=sig_kwargs,
        n_sig=n_sig,
        n_bkg_events_list=n_bkg_events_list,
        n_sig_events_list=n_sig_events_list,
        bkg_events_list=bkg_events_list,
        sig_events_list=sig_events_list
    )

    with TaskTimer(tl, 'Writing pseudo data to file.'):
        with open(filename, 'wb') as fp:
            pickle.dump(trial_data, fp)


def load_pseudo_data(filename, tl=None):
    """Loads the pseudo data for a single trial from the given file name.

    Parameters
    ----------
    filename : str
        The name of the file that contains the pseudo data.
    tl : TimeLord | None
        The instance of TimeLord that should be used to time individual tasks.

    Returns
    -------
    mean_n_sig : float
        The mean number of signal events that was used to generate the pseudo
        data.
    n_sig : int
        The actual total number of signal events in the pseudo data.
    n_bkg_events_list : list of int
        The total number of background events for each data set of the
        pseudo data.
    n_sig_events_list : list of int
        The total number of signal events for each data set of the pseudo data.
    bkg_events_list : list of DataFieldRecordArray instances
        The list of DataFieldRecordArray instances containing the background
        pseudo data events for each data set.
    sig_events_list : list of DataFieldRecordArray instances | None
        The list of DataFieldRecordArray instances containing the signal
        pseudo data events for each data set. If a particular dataset has
        no signal events, the entry for that dataset can be None.
    """
    with TaskTimer(tl, 'Loading pseudo data from file.'):
        with open(filename, 'rb') as fp:
            trial_data = pickle.load(fp)

    return (
        trial_data['mean_n_sig'],
        trial_data['n_sig'],
        trial_data['n_bkg_events_list'],
        trial_data['n_sig_events_list'],
        trial_data['bkg_events_list'],
        trial_data['sig_events_list']
    )



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.scrambling import (
    DataScramblingMethod,
    TimeScramblingMethod,
)

from skyllh.i3.utils.coords import (
    azi_to_ra_transform,
    hor_to_equ_transform,
)


class I3TimeScramblingMethod(
        TimeScramblingMethod,
):
    """The I3TimeScramblingMethod class provides a data scrambling method to
    perform time scrambling of the data,
    by drawing a MJD time from a given time generator.
    """
    def __init__(
            self,
            timegen,
            **kwargs,
    ):
        """Initializes a new I3 time scrambling instance.

        Parameters
        ----------
        timegen : TimeGenerator
            The time generator that should be used to generate random MJD times.
        """
        super().__init__(
            timegen=timegen,
            hor_to_equ_transform=hor_to_equ_transform,
            **kwargs)

    # We override the scramble method because for IceCube we only need to change
    # the ``ra`` field.
    def scramble(
            self,
            rss,
            dataset,
            data,
    ):
        """Draws a time from the time generator and calculates the right
        ascention coordinate from the azimuth angle according to the time.
        Sets the values of the ``time`` and ``ra`` keys of data.

        Parameters
        ----------
        rss : RandomStateService
            The random state service providing the random number
            generator (RNG).
        dataset : instance of Dataset
            The instance of Dataset for which the data should get scrambled.
        data : DataFieldRecordArray instance
            The DataFieldRecordArray instance containing the to be scrambled
            data.

        Returns
        -------
        data : numpy record ndarray
            The given numpy record ndarray holding the scrambled data.
        """
        mjds = self._timegen.generate_times(rss, len(data))

        data['time'] = mjds
        data['ra'] = azi_to_ra_transform(data['azi'], mjds)

        return data


class I3SeasonalVariationTimeScramblingMethod(
        DataScramblingMethod,
):
    """The I3SeasonalVariationTimeScramblingMethod class provides a data
    scrambling method to perform data coordinate scrambling based on a generated
    time, which follows seasonal variations within the experimental data.
    """
    def __init__(
            self,
            data,
            **kwargs,
    ):
        """Initializes a new seasonal time scrambling instance.

        Parameters
        ----------
        data : instance of I3DatasetData
            The instance of I3DatasetData holding the experimental data and
            good-run-list information.
        """
        super().__init__(**kwargs)

        # The run weights are the number of events in each run relative to all
        # the events to account for possible seasonal variations.
        self.run_weights = np.zeros((len(data.grl),), dtype=np.float64)
        n_events = len(data.exp['time'])
        for (i, (start, stop)) in enumerate(
                zip(data.grl['start'], data.grl['stop'])):
            mask = (data.exp['time'] >= start) & (data.exp['time'] < stop)
            self.run_weights[i] = len(data.exp[mask]) / n_events
        self.run_weights /= np.sum(self.run_weights)

        self.grl = data.grl

    def scramble(
            self,
            rss,
            dataset,
            data,
    ):
        """Scrambles the given data based on random MJD times, which are
        generated uniformely within the data runs, where the data runs are
        weighted based on their amount of events compared to the total events.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number
            generator (RNG).
        dataset : instance of Dataset
            The instance of Dataset for which the data should get scrambled.
        data : instance of DataFieldRecordArray
            The DataFieldRecordArray instance containing the to be scrambled
            data.

        Returns
        -------
        data : instance of DataFieldRecordArray
            The given DataFieldRecordArray holding the scrambled data.
        """
        # Get run indices based on their seasonal weights.
        run_idxs = rss.random.choice(
            self.grl['start'].size,
            size=len(data['time']),
            p=self.run_weights)

        # Draw random times uniformely within the runs.
        times = rss.random.uniform(
            self.grl['start'][run_idxs],
            self.grl['stop'][run_idxs])

        # Get the correct right ascension.
        data['time'] = times
        data['ra'] = azi_to_ra_transform(
            azi=data['azi'],
            mjd=times)

        return data



# -*- coding: utf-8 -*-

"""This module provides functionality for interpolation.
"""

import abc
import numpy as np

from skyllh.core.parameters import (
    ParameterGrid,
    ParameterGridSet
)
from skyllh.core.py import (
    classname,
)


class GridManifoldInterpolationMethod(
        object,
        metaclass=abc.ABCMeta,
):
    """This is an abstract base class for implementing a method to interpolate
    a manifold that is defined on a grid of parameter values. In general the
    number of parameters can be arbitrary and hence the manifold's
    dimensionality can be arbitrary, too. However, in practice the interpolation
    on a multi-dimensional manifold can be rather difficult.
    Nevertheless, we provide this interface to allow for manifold grids with
    different dimensionality.
    """

    def __init__(
            self,
            func,
            param_grid_set,
            **kwargs,
    ):
        """Constructor for a GridManifoldInterpolationMethod object.
        It must be called by the derived class.

        Parameters
        ----------
        func : callable R^D -> R
            The function that takes D parameter grid values as input and returns
            the value of the D-dimensional manifold at this point for each given
            trial event and source.
            The call signature of func must be:

                ``__call__(tdm, eventdata, gridparams_recarray, n_values, **kwargs)``

            The arguments are as follows:

                tdm : instance of TrialDataManager
                    The TrialDataManager instance holding the trial event data.
                eventdata : instance of numpy ndarray
                    A two-dimensional (V,N_events)-shaped numpy ndarray holding
                    the event data, where N_events is the number of trial
                    events, and V the dimensionality of the event data.
                gridparams_recarray : instance of numpy.ndarray
                    The structured numpy ndarray of length ``len(src_idxs)``
                    with the D parameter names and values on the grid for all
                    sources.
                n_values : int
                    The length of the output numpy ndarray of shape (n_values,).
                **kwargs
                    Additional keyword arguments required by ``func``.

            The return value of ``func`` should be the (n_values,)-shaped
            one-dimensional ndarray holding the values for each set of parameter
            values of the sources given via the ``gridparams_recarray``.
            The length of the array, i.e. n_values, depends on the
            ``src_evt_idx`` property of the TrialDataManager. In the worst case
            n_values is N_sources * N_events.
        param_grid_set : instance of ParameterGrid | instance of ParameterGridSet
            The set of D parameter grids. This defines the grid of the
            manifold.
        """
        super().__init__(**kwargs)

        self.func = func
        self.param_grid_set = param_grid_set

    @property
    def func(self):
        """The R^d -> R manifold function.
        """
        return self._func

    @func.setter
    def func(self, f):
        if not callable(f):
            raise TypeError(
                'The func property must be a callable object!')
        self._func = f

    @property
    def param_grid_set(self):
        """The ParameterGridSet instance defining the set of D parameter grids.
        This defines the grid of the manifold.
        """
        return self._param_grid_set

    @param_grid_set.setter
    def param_grid_set(self, obj):
        if isinstance(obj, ParameterGrid):
            obj = ParameterGridSet([obj])
        elif not isinstance(obj, ParameterGridSet):
            raise TypeError(
                'The param_grid_set property must be an instance '
                'of ParameterGrid or ParameterGridSet!')
        self._param_grid_set = obj

    @property
    def ndim(self):
        """(read-only) The dimensionality of the manifold.
        """
        return len(self._param_grid_set)

    @abc.abstractmethod
    def __call__(
            self,
            tdm,
            eventdata,
            params_recarray,
            **kwargs,
    ):
        """Retrieves the interpolated value of the manifold at the D-dimensional
        point ``params_recarray`` for all given events and sources, along with
        the D gradients, i.e. partial derivatives.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : numpy ndarray
            The 2D (V,N_events)-shaped numpy ndarray holding the event data,
            where N_events is the number of trial events, and V the
            dimensionality of the event data.
        params_recarray : instance of numpy record ndarray
            The structured numpy ndarray holding the N_sources set of parameter
            names and values, that define the point (for each source) on the
            manifold for which the value should get calculated for each event.
        **kwargs
            Additional keyword arguments required by the function ``func``.

        Returns
        -------
        values : instance of numpy.ndarray
            The (N,)-shaped numpy ndarray holding the interpolated manifold
            values for the given events and sources.
        grads : instance of numpy.ndarray
            The (D,N)-shaped numpy ndarray holding the D manifold gradients for
            the N given values, where D is the number of parameters.
            The order of the D parameters is defined by the ParameterGridSet
            that has been provided at construction time of this interpolation
            method object.
        """
        pass


class NullGridManifoldInterpolationMethod(
        GridManifoldInterpolationMethod,
):
    """This grid manifold interpolation method performs no interpolation. When
    the
    :meth:`~skyllh.core.interpolate.NullGridManifoldInterpolationMethod.__call__`
    method is called, it rounds the parameter values to their nearest grid
    point values. All gradients are set to zero.
    """

    def __init__(
            self,
            func,
            param_grid_set,
            **kwargs,
    ):
        """Creates a new NullGridManifoldInterpolationMethod instance.

        Parameters
        ----------
        func : callable R^d -> R
            The function that takes d parameter grid values as input and returns
            the value of the d-dimensional manifold at this point for each given
            trial event and source.
            See the documentation of the
            :class:`~skyllh.core.interpolate.GridManifoldInterpolationMethod`
            class for more details.
        param_grid_set : instance of ParameterGrid | instance of ParameterGridSet
            The set of d parameter grids. This defines the grid of the
            manifold.
        """
        super().__init__(
            func=func,
            param_grid_set=param_grid_set,
            **kwargs)

    def __call__(
            self,
            tdm,
            eventdata,
            params_recarray,
            **kwargs,
    ):
        """Calculates the non-interpolated manifold value and its gradient
        (zero) for each given event and source at the points given by
        ``params_recarray``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : instance of numpy.ndarray
            The (V,N_events)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params_recarray : instance of numpy.ndarray
            The structured numpy ndarray of length N_sources holding the
            parameter names and values of the sources, defining the point on the
            manifold for which the values should get calculated.
        **kwargs
            Additional keyword arguments required by the function ``func``.

        Returns
        -------
        values : instance of numpy.ndarray
            The (N,)-shaped numpy ndarray holding the interpolated manifold
            values for the given events and sources.
        grads : instance of numpy.ndarray
            The (D,N)-shaped ndarray of float holding the D manifold gradients
            for the N values, where D is the number of parameters of the
            manifold.
            By definition, all gradients are zero.
        """
        # Round the given parameter values to their nearest grid values.
        gridparams_recarray_dtype = [
            (p_grid.name, np.float64)
            for p_grid in self._param_grid_set
        ]

        gridparams_recarray = np.empty(
            params_recarray.shape,
            dtype=gridparams_recarray_dtype)

        for p_grid in self._param_grid_set:
            pname = p_grid.name
            pvalues = params_recarray[pname]
            gridparams_recarray[pname] = p_grid.round_to_nearest_grid_point(
                pvalues)

        values = self._func(
            tdm=tdm,
            eventdata=eventdata,
            gridparams_recarray=gridparams_recarray,
            n_values=tdm.get_n_values(),
            **kwargs)

        grads = np.zeros(
            (len(self.param_grid_set), len(values)),
            dtype=np.float64)

        return (values, grads)


class Linear1DGridManifoldInterpolationMethod(
        GridManifoldInterpolationMethod,
):
    """This grid manifold interpolation method interpolates the 1-dimensional
    grid manifold using a line.
    """

    def __init__(
            self,
            func,
            param_grid_set,
            **kwargs,
    ):
        """Creates a new Linear1DGridManifoldInterpolationMethod instance.

        Parameters
        ----------
        func : callable R -> R
            The function that takes the parameter grid value as input and
            returns the value of the 1-dimensional manifold at this point for
            each given source and trial event.
            See the documentation of the
            :class:`~skyllh.core.interpolate.GridManifoldInterpolationMethod`
            class for more details.
        param_grid_set : instance of ParameterGrid | instance of ParameterGridSet
            The one parameter grid. This defines the grid of the manifold.
        """
        super().__init__(
            func=func,
            param_grid_set=param_grid_set,
            **kwargs)

        if len(self._param_grid_set) != 1:
            raise ValueError(
                f'The {classname(self)} class supports only 1D grid manifolds. '
                'The param_grid_set argument must contain 1 ParameterGrid '
                f'instance! Currently it has {len(self._param_grid_set)}!')
        self._p_grid = self._param_grid_set[0]

        # Create a cache for the line parameterization for the last
        # manifold grid point for the different events.
        self._cache = self._create_cache(
            trial_data_state_id=None,
            x0=None,
            m=None,
            b=None)

    def _create_cache(
            self,
            trial_data_state_id,
            x0,
            m,
            b,
    ):
        """Creates a cache for the line parameterization for the last manifold
        grid point for the N_events different events.

        Parameters
        ----------
        trial_data_state_id : int | None
            The trial data state id of the TrialDataManager.
        x0 : instance of ndarray | None
            The (N_sources,)-shaped numpy ndarray holding the parameter grid
            value of the lower point of the grid manifold for each source used
            to estimate the line.
        m : instance of ndarray | None
            The (N_values,)-shaped numpy ndarray holding the slope of the line
            for each trial event and source.
        b : instance of ndarray | None
            The (N_values,)-shaped numpy ndarray holding the offset coefficient
            of the line for each trial event and source.
        """
        cache = {
            'trial_data_state_id': trial_data_state_id,
            'x0': x0,
            'm': m,
            'b': b
        }

        return cache

    def _is_cached(
            self,
            trial_data_state_id,
            x0,
    ):
        """Checks if the given line parametrization are already cached for the
        given x0 values.

        Returns
        -------
        check : bool
            ``True`` if the line parametrization for x0 is already cached,
            ``False`` otherwise.
        """
        self__cache = self._cache
        if (self__cache['trial_data_state_id'] is not None) and\
           (self__cache['trial_data_state_id'] == trial_data_state_id) and\
           (np.all(np.isclose(self__cache['x0'], x0))):
            return True

        return False

    def __call__(
            self,
            tdm,
            eventdata,
            params_recarray,
            **kwargs,
    ):
        """Calculates the interpolated manifold value and its gradient for each
        given source and trial event at the point ``params_recarray``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : instance of numpy ndarray
            The (V,N_events)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params_recarray : numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values for each source, defining the point on the manifold
            for which the value should get calculated.
            This record ndarray can be of length 1. In that case the single set
            of parameters is used for all sources.
        **kwargs
            Additional keyword arguments required by the function ``func``.

        Returns
        -------
        values : instance of numpy.ndarray
            The (N_values,)-shaped numpy ndarray of float holding the
            interpolated manifold values for all sources and trial events.
        grads : instance of numpy.ndarray
            The (D,N_values)-shaped numpy ndarray of float holding the D
            manifold gradients for the N_values values for all sources and trial
            events, where D is the number of interpolation parameters.
        """
        xname = self._p_grid.name

        x = params_recarray[xname]

        # Determine the nearest grid point that is lower than x and use that as
        # x0.
        x0 = self._p_grid.round_to_lower_grid_point(x)

        # Check if the line parametrization for x0 is already cached.
        if self._is_cached(tdm.trial_data_state_id, x0):
            m = self._cache['m']
            b = self._cache['b']

            (x,) = tdm.broadcast_sources_arrays_to_values_arrays((x,))

            values = m*x + b

            return (values, np.atleast_2d(m))

        # The line parametrization is not cached.
        # Calculate the line parametrization for all the given events.
        self__func = self._func

        x1 = self._p_grid.round_to_upper_grid_point(x)

        n_values = tdm.get_n_values()

        values = np.empty((n_values,), dtype=np.float64)
        m = np.empty((n_values,), dtype=np.float64)

        gridparams_recarray = np.array(
            x0,
            dtype=[(xname, np.float64)])
        M0 = self__func(
            tdm=tdm,
            eventdata=eventdata,
            gridparams_recarray=gridparams_recarray,
            n_values=n_values,
            **kwargs)

        gridparams_recarray = np.array(
            x1,
            dtype=[(xname, np.float64)])
        M1 = self__func(
            tdm=tdm,
            eventdata=eventdata,
            gridparams_recarray=gridparams_recarray,
            n_values=n_values,
            **kwargs)

        # Broadcast x0 and x1 to the values array.
        (x, v_x0, v_x1) = tdm.broadcast_sources_arrays_to_values_arrays(
            (x, x0, x1))

        m = (M1 - M0) / (v_x1 - v_x0)
        b = M0 - m*v_x0

        # Cache the line parametrization.
        self._cache = self._create_cache(
            trial_data_state_id=tdm.trial_data_state_id,
            x0=x0,
            m=m,
            b=b)

        # Calculate the interpolated manifold values. The gradient is m.
        values = m*x + b

        return (values, np.atleast_2d(m))


class Parabola1DGridManifoldInterpolationMethod(
        GridManifoldInterpolationMethod,
):
    """This grid manifold interpolation method interpolates the 1-dimensional
    grid manifold using a parabola.
    """

    def __init__(
            self,
            func,
            param_grid_set,
            **kwargs,
    ):
        """Creates a new Parabola1DGridManifoldInterpolationMethod instance.

        Parameters
        ----------
        func : callable R -> R
            The function that takes the parameter grid value as input and
            returns the value of the 1-dimensional manifold at this point for
            each given source and trial event.
            See the documentation of the
            :class:`~skyllh.core.interpolate.GridManifoldInterpolationMethod`
            class for more details.
        param_grid_set : instance of ParameterGrid | instance of ParameterGridSet
            The one parameter grid. This defines the grid of the manifold.
        """
        super().__init__(
            func=func,
            param_grid_set=param_grid_set,
            **kwargs)

        if len(self._param_grid_set) != 1:
            raise ValueError(
                f'The {classname(self)} class supports only 1D grid manifolds. '
                'The param_grid_set argument must contain 1 ParameterGrid '
                f'instance! Currently it has {len(self._param_grid_set)}!')
        self._p_grid = self._param_grid_set[0]

        # Create a cache for the parabola parameterization for the last
        # manifold grid point for the different events.
        self._cache = self._create_cache(
            trial_data_state_id=None,
            x1=None,
            M1=None,
            a=None,
            b=None)

    def _create_cache(
            self,
            trial_data_state_id,
            x1,
            M1,
            a,
            b,
    ):
        """Creates a cache for the parabola parameterization for the last
        manifold grid point for the N_events different events.

        Parameters
        ----------
        trial_data_state_id : int | None
            The trial data state ID of the TrialDataManager.
        x1 : instance of numpy.ndarray | None
            The (N_sources,)-shaped numpy ndarray of float holding the parameter
            grid value for the middle point of the grid manifold for all sources
            used to estimate the parabola.
        M1 : instance of numpy.ndarray
            The (N_values,)-shaped numpy ndarray of float holding the grid
            manifold value for each source and trial event of the middle point
            (x1,).
        a : instance of numpy.ndarray
            The (N_values,)-shaped numpy ndarray of float holding the parabola
            coefficient ``a`` for each source and trial event.
        b : instance of numpy.ndarray
            The (N_values,)-shaped numpy ndarray of float holding the parabola
            coefficient ``b`` for each source and trial event.

        Returns
        -------
        cache : dict
            The dictionary holding the cache data.
        """
        cache = {
            'trial_data_state_id': trial_data_state_id,
            'x1': x1,
            'M1': M1,
            'a': a,
            'b': b,
        }

        return cache

    def _is_cached(
            self,
            trial_data_state_id,
            x1,
    ):
        """Checks if the parabola parametrization is already cached for the
        given x1 values.
        """
        self__cache = self._cache
        if (self__cache['trial_data_state_id'] is not None) and\
           (self__cache['trial_data_state_id'] == trial_data_state_id):
            if np.any(np.not_equal(self__cache['x1'], x1)):
                return False
            return True

        return False

    def __call__(
            self,
            tdm,
            eventdata,
            params_recarray,
            **kwargs,
    ):
        """Calculates the interpolated manifold value and its gradient for each
        given source and trial event at the point ``params_recarray``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial data.
        eventdata : instance of numpy ndarray
            The (V,N_events)-shaped numpy ndarray holding the event data,
            where N_events is the number of events, and V the dimensionality of
            the event data.
        params_recarray : numpy record ndarray
            The numpy record ndarray of length N_sources holding the parameter
            names and values for each source, defining the point on the manifold
            for which the value should get calculated.
            This record ndarray can be of length 1. In that case the single set
            of parameters is used for all sources.
        **kwargs
            Additional keyword arguments required by the function ``func``.

        Returns
        -------
        values : (N_values,) ndarray of float
            The interpolated manifold value for the N given events.
        grads : (D,N_values) ndarray of float
            The D manifold gradients for the N given events, where D is the
            number of parameters.
        """
        xname = self._p_grid.name

        x = params_recarray[xname]

        # Determine the nearest grid point x1.
        x1 = self._p_grid.round_to_nearest_grid_point(x)

        # Check if the parabola parametrization for x1 is already cached.
        if self._is_cached(tdm.trial_data_state_id, x1):
            M1 = self._cache['M1']
            a = self._cache['a']
            b = self._cache['b']
        else:
            dx = self._p_grid.delta

            # Calculate the neighboring grid points to x1: x0 and x2.
            x0 = self._p_grid.round_to_nearest_grid_point(x1 - dx)
            x2 = self._p_grid.round_to_nearest_grid_point(x1 + dx)

            # Parameterize the parabola with parameters a, b, and M1.
            self__func = self._func

            n_values = tdm.get_n_values()

            gridparams_recarray = np.array(
                x0,
                dtype=[(xname, np.float64)])
            M0 = self__func(
                tdm=tdm,
                eventdata=eventdata,
                gridparams_recarray=gridparams_recarray,
                n_values=n_values,
                **kwargs)

            gridparams_recarray = np.array(
                x1,
                dtype=[(xname, np.float64)])
            M1 = self__func(
                tdm=tdm,
                eventdata=eventdata,
                gridparams_recarray=gridparams_recarray,
                n_values=n_values,
                **kwargs)

            gridparams_recarray = np.array(
                x2,
                dtype=[(xname, np.float64)])
            M2 = self__func(
                tdm=tdm,
                eventdata=eventdata,
                gridparams_recarray=gridparams_recarray,
                n_values=n_values,
                **kwargs)

            a = 0.5*(M0 - 2.*M1 + M2) / dx**2
            b = 0.5*(M2 - M0) / dx

            # Cache the parabola parametrization.
            self._cache = self._create_cache(
                trial_data_state_id=tdm.trial_data_state_id,
                x1=x1,
                M1=M1,
                a=a,
                b=b)

        # Broadcast x, x1, and (x-x1) to the values array.
        (x, x1, x_minus_x1) = tdm.broadcast_sources_arrays_to_values_arrays(
            (x, x1, x-x1))

        # Calculate the interpolated manifold values.
        values = a * x_minus_x1**2 + b * x_minus_x1 + M1
        # Calculate the gradient of the manifold for all values.
        grads = 2. * a * x_minus_x1 + b

        return (values, np.atleast_2d(grads))



# -*- coding: utf-8 -*-

"""IceCube specific coordinate utility functions.
"""

import numpy as np


def azi_to_ra_transform(azi, mjd):
    """Rotates the given IceCube azimuth angles into right-ascention angles for
    the given MJD times. This function is IceCube specific and assumes that the
    detector is located excently at the South Pole and neglects all astronomical
    effects like Earth's precession.

    Parameters
    ----------
    azi : instance of numpy.ndarray
        The array with the azimuth angles.
    mjd : instance of numpy.ndarray
        The array with the MJD times for each azimuth angle.

    Returns
    -------
    ra : instance of numpy.ndarray
        The right-ascention values.
    """
    # sidereal day = length * solar day
    _sidereal_length = 0.997269566
    _sidereal_offset = 2.54199002505
    sidereal_day_residuals = (mjd / _sidereal_length) % 1
    ra = _sidereal_offset + 2 * np.pi * sidereal_day_residuals - azi
    ra = np.mod(ra, 2*np.pi)

    return ra


def ra_to_azi_transform(ra, mjd):
    """Rotates the given right-ascention angles to local IceCube azimuth angles.

    Parameters
    ----------
    ra : instance of numpy.ndarray
        The array with the right-ascention angles.
    mjd : instance of numpy.ndarray
        The array with the MJD times for each right-ascention angle.

    Returns
    -------
    azi : instance of numpy.ndarray
        The azimuth angle for each right-ascention angle.
    """
    # Use the azi_to_ra_transform function because it is symmetric.
    azi = azi_to_ra_transform(ra, mjd)

    return azi


def hor_to_equ_transform(azi, zen, mjd):
    """Transforms the coordinate from the horizontal system (azimuth, zenith)
    into the equatorial system (right-ascention, declination) for detector at
    the South Pole and neglecting all astronomical effects like Earth
    precession.

    Parameters
    ----------
    azi : instance of numpy.ndarray
        The azimuth angle.
    zen : instance of numpy.ndarray
        The zenith angle.
    mjd : instance of numpy.ndarray
        The time in MJD.

    Returns
    -------
    ra : instance of numpy.ndarray
        The right-ascention angle.
    dec : instance of numpy.ndarray
        The declination angle.
    """
    ra = azi_to_ra_transform(azi, mjd)
    dec = np.pi - zen
    return (ra, dec)



# -*- coding: utf-8 -*-

import abc

from skyllh.core.livetime import Livetime


class TimeGenerationMethod(
        object,
        metaclass=abc.ABCMeta,
):
    """Base class (type) for implementing a method to generate times.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def generate_times(
            self,
            rss,
            size,
    ):
        """The ``generate_times`` method implements the actual generation of
        times, which is method dependent.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number
            generator (RNG).
        size : int
            The number of times that should get generated.

        Returns
        -------
        times : ndarray
            The 1d numpy ndarray holding the generated times.
        """
        pass


class LivetimeTimeGenerationMethod(
        TimeGenerationMethod,
):
    """The LivetimeTimeGenerationMethod provides the method to generate times
    from a Livetime object. It will uniformely generate times that will coincide
    with the on-time intervals of the detector, by calling the `draw_ontimes`
    method of the Livetime class.
    """
    def __init__(self, livetime, **kwargs):
        """Creates a new LivetimeTimeGeneration instance.

        Parameters
        ----------
        livetime : Livetime
            The Livetime instance that should be used to generate times from.
        """
        super().__init__(**kwargs)

        self.livetime = livetime

    @property
    def livetime(self):
        """The Livetime instance used to draw times from.
        """
        return self._livetime

    @livetime.setter
    def livetime(self, livetime):
        if not isinstance(livetime, Livetime):
            raise TypeError(
                'The livetime property must be an instance of Livetime!')
        self._livetime = livetime

    def generate_times(
            self,
            rss,
            size,
            **kwargs,
    ):
        """Generates `size` MJD times according to the detector on-times
        provided by the Livetime instance.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number
            generator (RNG).
        size : int
            The number of times that should get generated.

        Returns
        -------
        times : ndarray
            The 1d (`size`,)-shaped numpy ndarray holding the generated times.
        """
        times = self._livetime.draw_ontimes(
            rss=rss,
            size=size,
            **kwargs)

        return times


class TimeGenerator(
        object):
    def __init__(self, method):
        """Creates a time generator instance with a given defined time
        generation method.

        Parameters
        ----------
        method : instance of TimeGenerationMethod
            The instance of TimeGenerationMethod that defines the method of
            generating times.
        """
        self.method = method

    @property
    def method(self):
        """The TimeGenerationMethod object that should be used to generate
        the times.
        """
        return self._method

    @method.setter
    def method(self, method):
        if not isinstance(method, TimeGenerationMethod):
            raise TypeError(
                'The time generation method must be an instance of '
                'TimeGenerationMethod!')
        self._method = method

    def generate_times(
            self,
            rss,
            size,
            **kwargs,
    ):
        """Generates ``size`` amount of times by calling the ``generate_times``
        method of the TimeGenerationMethod class.

        Parameters
        ----------
        rss : instance of RandomStateService
            The random state service providing the random number generator
            (RNG).
        size : int
            The number of time that should get generated.
        **kwargs
            Additional keyword arguments are passed to the ``generate_times``
            method of the TimeGenerationMethod class.

        Returns
        -------
        times : ndarray
            The 1d (``size``,)-shaped ndarray holding the generated times.
        """
        times = self._method.generate_times(
            rss=rss,
            size=size,
            **kwargs)

        return times



# -*- coding: utf-8 -*-

"""The tool module provides functionality to interface with an optional external
python package (tool). The tool can be imported dynamically at run-time when
needed.
"""

import importlib
import importlib.util
import sys

import numpy as np

from skyllh.core.py import (
    classname,
    get_class_of_func,
)


def assert_tool_version(
        tool,
        version,
):
    """Asserts the required version of the tool. The tool module must have the
    attribute ``__version__``.

    Parameters
    ----------
    tool : str
        The name of the tool.
    version : str
        The required version of the tool in the format ``"<COMP>X.Y.Z"``, where
        ``<COMP>`` is one of ``<=``, ``==``, and ``>=``.

    Raises
    ------
    ImportError
        If the version of the tool does not match the requirements.
    KeyError
        If the tool module has no attribute named ``__version__``.
    ValueError
        If ``<COMP>`` is not supported.
    """
    tool_module = get(tool)
    if not hasattr(tool_module, '__version__'):
        raise KeyError(
            f'The tool "{tool}" has no attribute "__version__"!')
    tool_version_arr = tool_module.__version__.split('.')
    (comp_op, version) = (version[0:2], version[2:])
    version_arr = version.split('.')
    for i in range(np.min([len(tool_version_arr), len(version_arr)])):
        tool_vers_i = int(tool_version_arr[i])
        vers_i = int(version_arr[i])
        if comp_op == '<=':
            if not (tool_vers_i <= vers_i):
                raise ImportError(
                    f'The version ({".".join(tool_version_arr)}) of the tool '
                    f'"{tool}" is not lower or equal than version '
                    f'{".".join(version_arr)}!')
        elif comp_op == '==':
            if not (tool_vers_i == vers_i):
                raise ImportError(
                    f'The version ({".".join(tool_version_arr)}) of the tool '
                    f'"{tool}" is not equal to the version '
                    f'{".".join(version_arr)}!')
        elif comp_op == '>=':
            if not (tool_vers_i >= vers_i):
                raise ImportError(
                    f'The version ({".".join(tool_version_arr)}) of the tool '
                    f'"{tool}" is not greater or equal than version '
                    f'{".".join(version_arr)}!')
        else:
            raise ValueError(
                f'The version comparison operator "{comp_op}" for the tool '
                f'"{tool}" is not supported!')


def is_available(name):
    """Checks if the given Python package is available for import.

    Parameters
    ----------
    name : str
        The name of the Python package.

    Returns
    -------
    check : bool
        ``True`` if the given Python package is available, ``False`` otherwise.

    Raises
    ------
    ModuleNotFoundError
        If the package is not a Python package, i.e. lacks a __path__ attribute.
    """
    # Check if module is already imported.
    if name in sys.modules:
        return True

    spec = importlib.util.find_spec(name)
    if spec is not None:
        return True

    return False


def get(name):
    """Returns the module object of the given tool. This will import the Python
    package if it was not yet imported.

    Parameters
    ----------
    name : str
        The name of the Python package.

    Returns
    -------
    module : Python module
        The (imported) Python module object.
    """
    if name in sys.modules:
        return sys.modules[name]

    module = importlib.import_module(name)
    return module


def _get_tool_and_version(
        tool,
):
    """Returns the tool and and version based on the input value for the tool.

    Parameters
    ----------
    tool : str | (str, str)
        Either the tool name or the tuple with the tool name and required
        version string.

    Returns
    -------
    tool : str
        The name of the tool.
    version : str | None
        The tool's version string, or ``None``, if no version was specified.
    """
    if not (isinstance(tool, str) or isinstance(tool, tuple)):
        raise TypeError(
            'The tool specification must be an instance of str or tuple! '
            f'Its current type is {classname(tool)}!')

    if isinstance(tool, tuple):
        if len(tool) != 2:
            raise ValueError(
                'The length of the tool tuple must be 2, but it is '
                f'{len(tool)}!')

        (tool, version) = tool

        if not isinstance(tool, str):
            raise TypeError(
                'The first element of the tool tuple must be an instance '
                'of str! '
                f'Its current type is {classname(tool)}!')
        if not isinstance(version, str):
            raise TypeError(
                'The second element of the tool tuple must be an instance '
                'of str! '
                f'Its current type is {classname(version)}!')

        return (tool, version)

    return (tool, None)


def requires(*tools):
    """This is decorator function that can be used whenever a function requires
    optional tools.

    Parameters
    ----------
    *tools : sequence of str | sequence of (str, str)
        The name(s) of the required Python packages.
        If a 2-element tuple of (str, str) is provided for a tool, the first
        element specifies the name of the tool and the second element its
        version of the form ``"<COMP>X.Y.Z"``, where ``<COMP>`` is one of
        ``<=``, ``==``, and ``>=``.

    Raises
    ------
    ModuleNotFoundError
        If any of the specified tools is not available.
    ImportError
        If the version of a tool does not meet the requirements.
    """
    def decorator(f):
        def wrapper(*args, **kwargs):
            for tool in tools:
                (tool, version) = _get_tool_and_version(tool)
                if not is_available(tool):
                    raise ModuleNotFoundError(
                        f'The Python module "{tool}" is not available, but is '
                        f'required by "{get_class_of_func(f)}.{f.__name__}"!')
                if version is not None:
                    assert_tool_version(tool, version)

            return f(*args, **kwargs)
        return wrapper
    return decorator



# -*- coding: utf-8 -*-

"""The ``signalpdf`` module contains possible signal PDF models for the
likelihood function.
"""

import numpy as np

from skyllh.core import (
    tool,
)
from skyllh.core.debugging import (
    get_logger,
)
from skyllh.core.interpolate import (
    GridManifoldInterpolationMethod,
    Linear1DGridManifoldInterpolationMethod,
)
from skyllh.core.pdf import (
    PDF,
    PDFSet,
    IsSignalPDF,
    MultiDimGridPDF,
    SpatialPDF,
    TimePDF,
)
from skyllh.core.py import (
    classname,
    str_cast,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.timing import (
    TaskTimer,
)
from skyllh.core.utils.coords import (
    angular_separation,
)


class GaussianPSFPointLikeSourceSignalSpatialPDF(
        SpatialPDF,
        IsSignalPDF):
    r"""This spatial signal PDF model describes the spatial PDF for a point
    source smeared with a 2D gaussian point-spread-function (PSF).
    Mathematically, it's the convolution of a point in the sky, i.e. the source
    location, with the PSF. The result of this convolution has the gaussian form

    .. math::

        \frac{1}{2\pi\sigma^2} \exp(-\frac{r^2}{2\sigma^2}),

    where :math:`\sigma` is the spatial uncertainty of the event and :math:`r`
    the distance on the sphere between the source and the data event.

    This PDF requires the `src_array` data field, that is numpy record ndarray
    with the data fields `ra` and `dec` holding the right-ascension and
    declination of the point-like sources, respectively.
    """

    def __init__(
            self,
            ra_range=None,
            dec_range=None,
            pd_event_data_field_name=None,
            **kwargs):
        """Creates a new spatial signal PDF for point-like sources with a
        gaussian point-spread-function (PSF).

        Parameters
        ----------
        ra_range : 2-element tuple | None
            The range in right-ascension this spatial PDF is valid for.
            If set to None, the range (0, 2pi) is used.
        dec_range : 2-element tuple | None
            The range in declination this spatial PDF is valid for.
            If set to None, the range (-pi/2, +pi/2) is used.
        pd_event_data_field_name : str | None
            The probability density values can be pre-calculated by the user.
            This specifies the name of the event data field, where these values
            are stored.
        """
        if ra_range is None:
            ra_range = (0, 2*np.pi)
        if dec_range is None:
            dec_range = (-np.pi/2, np.pi/2)

        super().__init__(
            pmm=None,
            ra_range=ra_range,
            dec_range=dec_range,
            **kwargs)

        self.pd_event_data_field_name = pd_event_data_field_name

    @property
    def pd_event_data_field_name(self):
        """The event data field name where pre-calculated probability density
        values are stored.
        """
        return self._pd_event_data_field_name

    @pd_event_data_field_name.setter
    def pd_event_data_field_name(self, name):
        name = str_cast(
            name,
            'The pd_event_data_field_name property must be cast-able to type '
            f'str! Its current type is {classname(name)}!')
        self._pd_event_data_field_name = name

    def calculate_pd(self, tdm):
        """Calculates the gaussian PSF probability density values for all events
        and sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF values. The following data fields need to
            be present:

            src_array : numpy record ndarray
                The numpy record ndarray with the following data fields:

                ra : float
                    The right-ascension of the point-like source.
                dec : float
                    The declination of the point-like source.

            ra : float
                The right-ascension in radian of the data event.
            dec : float
                The declination in radian of the data event.
            ang_err: float
                The reconstruction uncertainty in radian of the data event.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            for each event. The length of this 1D array depends on the number
            of sources and the events belonging to those sources. In the worst
            case the length is N_sources * N_trial_events.
        """
        get_data = tdm.get_data

        src_array = get_data('src_array')
        ra = get_data('ra')
        dec = get_data('dec')
        sigma = get_data('ang_err')

        (src_idxs, evt_idxs) = tdm.src_evt_idxs
        src_ra = np.take(src_array['ra'], src_idxs)
        src_dec = np.take(src_array['dec'], src_idxs)

        dec = np.take(dec, evt_idxs)
        ra = np.take(ra, evt_idxs)
        sigma_sq = np.take(sigma**2, evt_idxs)

        psi = angular_separation(src_ra, src_dec, ra, dec)

        pd = 0.5/(np.pi*sigma_sq) * np.exp(-0.5*(psi**2/sigma_sq))

        return pd

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """Calculates the spatial signal probability density of each event for
        all sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF values. The following data fields need to
            be present:

            src_array : numpy record ndarray
                The numpy record ndarray with the following data fields:

                ra : float
                    The right-ascension of the point-like source.
                dec : float
                    The declination of the point-like source.

            ra : float
                The right-ascension in radian of the data event.
            dec : float
                The declination in radian of the data event.
            ang_err: float
                The reconstruction uncertainty in radian of the data event.

            In case the probability density values were pre-calculated,
        params_recarray : None
            Unused interface argument.
        tl : TimeLord instance | None
            The optional TimeLord instance to use for measuring timing
            information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            for each event. The length of this 1D array depends on the number
            of sources and the events belonging to those sources. In the worst
            case the length is N_sources * N_trial_events.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each fit parameter. By definition this PDF does not depend
            on any fit parameters and hence, this dictionary is empty.
        """

        logger = get_logger(f'{__name__}.{classname(self)}.get_pd')

        # Check if the probability density was pre-calculated.
        if self._pd_event_data_field_name in tdm:
            if self._cfg.is_tracing_enabled:
                logger.debug(
                    'Retrieve precalculated probability density values from '
                    f'data field "{self._pd_event_data_field_name}"')
            pd = tdm[self._pd_event_data_field_name]
            return (pd, dict())

        pd = self.calculate_pd(tdm)

        return (pd, dict())


class RayleighPSFPointSourceSignalSpatialPDF(
        SpatialPDF,
        IsSignalPDF):
    r"""This spatial signal PDF model describes the spatial PDF for a point-like
    source following a Rayleigh distribution in the opening angle between the
    source and reconstructed muon direction.
    Mathematically, it's the convolution of a point in the sky, i.e. the source
    location, with the PSF. The result of this convolution has the following
    form

    .. math::

        1/(2\pi \sin\Psi) * \Psi/\sigma^2 \exp(-\Psi^2/(2\sigma^2)),

    where :math:`\sigma` is the spatial uncertainty of the event and
    :math:`\Psi` the distance on the sphere between the source and the data
    event.

    This PDF requires the ``src_array`` source data field, that is numpy
    structured ndarray with the data fields ``ra`` and ``dec`` holding the
    right-ascension and declination of the point-like sources, respectively.
    """
    def __init__(
            self,
            ra_range=None,
            dec_range=None,
            **kwargs):
        r"""Creates a new spatial signal PDF for point-like sources with a
        Rayleigh point-spread-function (PSF).

        Parameters
        ----------
        ra_range : 2-element tuple | None
            The range in right-ascension this spatial PDF is valid for.
            If set to None, the range (0, 2pi) is used.
        dec_range : 2-element tuple | None
            The range in declination this spatial PDF is valid for.
            If set to None, the range (-pi/2, +pi/2) is used.
        """
        if ra_range is None:
            ra_range = (0, 2*np.pi)
        if dec_range is None:
            dec_range = (-np.pi/2, np.pi/2)

        super().__init__(
            pmm=None,
            ra_range=ra_range,
            dec_range=dec_range,
            **kwargs
        )

        self._pd = None

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs):
        """Pre-computes the probability density values once a new trial data is
        available.
        """
        get_data = tdm.get_data

        (src_idxs, evt_idxs) = tdm.src_evt_idxs

        psi = get_data('psi')
        sigma = get_data('ang_err')
        sigma_sq = np.take(sigma**2, evt_idxs)

        self._pd = (
            0.5/(np.pi*np.sin(psi)) *
            (psi / sigma_sq) *
            np.exp(-0.5*(psi**2/sigma_sq))
        )

    def get_pd(
            self,
            tdm,
            params_recarray=None,
            tl=None):
        """Calculates the spatial signal probability density of each event for
        all sources.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF values. The following data fields need to
            be present:

            psi : float
                The opening angle in radian between the source direction and the
                reconstructed muon direction.
            ang_err: float
                The reconstruction uncertainty in radian of the data event.

        params_recarray : None
            Unused interface argument.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to use for measuring timing
            information.

        Returns
        -------
        pd : (N_values,)-shaped numpy ndarray
            The (N_values,)-shaped 1D numpy ndarray holding the probability
            density value for each event and source in unit 1/rad.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each global fit parameter. By definition this PDF does not
            depend on any global fit parameters and hence, this dictionary is
            empty.
        """
        if self._pd is None:
            raise ValueError(
                'The PDF has not been initialized with trial data!')

        grads = dict()

        return (self._pd, grads)


class SignalTimePDF(
        TimePDF,
        IsSignalPDF,
):
    """This class provides a signal time PDF class. It consists of
    a :class:`~skyllh.core.livetime.Livetime` instance and a
    :class:`~skyllh.core.flux_model.TimeFluxProfile` instance. Together they
    construct the actual signal time PDF, which has detector down-time taking
    into account.
    """

    def __init__(
            self,
            livetime,
            time_flux_profile,
            **kwargs
    ):
        """Creates a new signal time PDF instance for a given time flux profile
        and detector live time.

        Parameters
        ----------
        livetime : instance of Livetime
            An instance of Livetime, which provides the detector live-time
            information.
        time_flux_profile : instance of TimeFluxProfile
            The signal's time flux profile.

            .. note::

                This instance of TimeFluxProfile will be altered by this PDF
                class when calculating the probability density values!

        """
        super().__init__(
            livetime=livetime,
            time_flux_profile=time_flux_profile,
            **kwargs)

        self._pd = None

    def _calculate_pd(
            self,
            tdm,
            params_recarray,
            tl=None,
    ):
        """Calculates the probability density values for the given trial data
        and source parameters.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data.
            The following data fields must exist:

            ``'time'`` : float
                The time of the event.

        params_recarray : instance of structured ndarray
            The structured numpy ndarray of length N_sources holding the local
            parameter names and values of the sources.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            values for each trial data event and source.
        """
        (src_idxs, evt_idxs) = tdm.src_evt_idxs
        n_values = len(evt_idxs)

        pd = np.zeros((n_values,), dtype=np.float64)

        events_time = tdm.get_data('time')
        for (src_idx, src_params_row) in enumerate(params_recarray):
            params = dict(zip(
                params_recarray.dtype.fields.keys(),
                src_params_row))

            # Update the time flux profile if its parameter values have changed
            # and recalculate self._I and self._S if an update was actually
            # performed.
            updated = self._time_flux_profile.set_params(params)
            if updated:
                self._S = self._calculate_sum_of_ontime_time_flux_profile_integrals()

            src_m = src_idxs == src_idx
            idxs = evt_idxs[src_m]

            times = events_time[idxs]

            # Get a mask of the event times which fall inside a detector on-time
            # interval.
            on = self._livetime.is_on(times)

            pd_src = pd[src_m]
            pd_src[on] = (
                self._time_flux_profile(t=times[on]) / self._S
            )
            pd[src_m] = pd_src

        return pd

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs,
    ):
        # Check if this time PDF is not constant and does depend on any global
        # floating parameters. If that's not the case we can pre-calculate the
        # PDF values.
        is_constant = (
            (self.param_set is None) or
            (len(self.param_set.params_name_list) == 0) or
            ((self.pmm is not None) and
             np.all(~self.pmm.get_local_param_is_global_floating_param_mask(
                 self.param_set.params_name_list)))
        )
        if not is_constant:
            self._pd = None
            return

        # At this point it has been checked that the PDF is constant and we can
        # pre-calculate the PDF values.

        if self.pmm is None:
            params_recarray = np.empty((tdm.n_sources,), dtype=[])
        else:
            params_recarray = self.pmm.create_src_params_recarray()

        self._pd = self._calculate_pd(
            tdm=tdm,
            params_recarray=params_recarray,
            tl=tl)

    def get_pd(
            self,
            tdm,
            params_recarray,
            tl=None,
    ):
        """Calculates the signal time probability density of each event for the
        given set of time parameter values for each source.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data for
            which to calculate the PDF value. The following data fields must
            exist:

            ``'time'`` : float
                The time of the event.

        params_recarray : instance of numpy structured ndarray
            The numpy structured ndarray holding the local parameter values for
            each source.
        tl : instance of TimeLord | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_values,)-shaped 1D numpy ndarray holding the probability
            density value for each trial event and source.
        grads : dict
            The dictionary holding the gradients of the probability density
            w.r.t. each global fit parameter.
        """
        # Check if we have pre-calculated PDF values.
        if self._pd is not None:
            return (self._pd, dict())

        pd = self._calculate_pd(
            tdm=tdm,
            params_recarray=params_recarray,
            tl=tl)

        return (pd, dict())


class SignalMultiDimGridPDF(
        MultiDimGridPDF,
        IsSignalPDF):
    """This class provides a multi-dimensional signal PDF. The PDF is created
    from pre-calculated PDF data on a grid. The grid data is interpolated using
    a :class:`scipy.interpolate.RegularGridInterpolator` instance.
    """

    def __init__(
            self,
            *args,
            **kwargs):
        """Creates a new PDF instance for a multi-dimensional PDF given
        as PDF values on a grid or as PDF values stored in a photospline table.

        See the documentation of the
        :meth:`skyllh.core.pdf.MultiDimGridPDF.__init__` method for the
        documentation of possible arguments.
        """
        super().__init__(
            *args,
            **kwargs)


class SignalMultiDimGridPDFSet(
        IsSignalPDF,
        PDFSet,
        PDF,
):
    """This class provides a set of MultiDimGridPDF instances that implements
    also the PDF interface.
    """

    def __init__(
            self,
            pmm,
            param_set,
            param_grid_set,
            gridparams_pdfs,
            interpol_method_cls=None,
            use_same_photospline_bfi_for_all_pdfs=False,
            **kwargs,
    ):
        """Creates a new MultiDimGridPDFSet instance, which holds a set of
        MultiDimGridPDF instances, one for each point of a parameter grid set.

        Parameters
        ----------
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper that defines the mapping of
            the global parameters to local model parameters.
        param_set : instance of Parameter | sequence of instance of Parameter | instance of ParameterSet
            The set of parameters defining the parameters of this PDF.
        param_grid_set : ParameterGrid instance | ParameterGridSet instance
            The set of ParameterGrid instances, which define the grid values of
            the model parameters, the given MultiDimGridPDF instances belong to.
        gridparams_pdfs : sequence of (dict, MultiDimGridPDF) tuples
            The sequence of 2-element tuples which define the mapping of grid
            values to PDF instances.
        interpol_method_cls : subclass of GridManifoldInterpolationMethod
            The class specifying the interpolation method. This must be a
            subclass of ``GridManifoldInterpolationMethod``.
            If set to None, the default grid manifold interpolation method
            ``Linear1DGridManifoldInterpolationMethod`` will be used.
        use_same_photospline_bfi_for_all_pdfs : bool
            Flag if the same basis function indices (bfi) should be used for
            all PDFs when photospline tables are used. Default is ``False``.
            This should be set to ``True`` if all photospline tables share the
            same binning. In that case the PDF evaluation will be accelerated.
        """
        super().__init__(
            pmm=pmm,
            param_set=param_set,
            param_grid_set=param_grid_set,
            **kwargs)

        if interpol_method_cls is None:
            interpol_method_cls = Linear1DGridManifoldInterpolationMethod
        self.interpol_method_cls = interpol_method_cls

        self.use_same_photospline_bfi_for_all_pdfs =\
            use_same_photospline_bfi_for_all_pdfs

        # Add the given MultiDimGridPDF instances to the PDF set.
        for (gridparams, pdf) in gridparams_pdfs:
            self.add_pdf(pdf, gridparams)

        # Create the interpolation method instance.
        self._interpol_method = self._interpol_method_cls(
            func=self._evaluate_pdfs,
            param_grid_set=self.param_grid_set)

        # Save the parameter names needed for the interpolation for later usage.
        self._interpol_param_names =\
            self.param_grid_set.params_name_list

        self._cache_eventdata = None

        # Determine if the PDFs are internally represented as
        # photospline.SplineTable instances.
        self.uses_photospline_SplineTable = False
        pdf = next(iter(self.items()))[1]
        if tool.is_available('photospline') and\
           isinstance(pdf.pdf, tool.get('photospline').SplineTable):
            self.uses_photospline_SplineTable = True

    @property
    def interpol_method_cls(self):
        """The class derived from GridManifoldInterpolationMethod
        implementing the interpolation of the PDF grid manifold.
        """
        return self._interpol_method_cls

    @interpol_method_cls.setter
    def interpol_method_cls(self, cls):
        if not issubclass(cls, GridManifoldInterpolationMethod):
            raise TypeError(
                'The interpol_method_cls property must be a sub-class of '
                'GridManifoldInterpolationMethod!')
        self._interpol_method_cls = cls

    def _get_pdf_for_interpol_param_values(
            self,
            interpol_param_values,
    ):
        """Retrieves the PDF for the given set of interpolation parameter
        values.

        Parameters
        ----------
        interpol_param_values : instance of numpy ndarray
            The (N_interpol_params,)-shaped numpy ndarray holding the values of
            the interpolation parameters.

        Returns
        -------
        pdf : instance of MultiDimGridPDF
            The requested PDF instance.
        """
        gridparams = dict(
            zip(self._interpol_param_names, interpol_param_values))

        pdf = self.get_pdf(gridparams)

        return pdf

    def _evaluate_pdfs(
            self,
            tdm,
            eventdata,
            gridparams_recarray,
            n_values,
            tl=None,
    ):
        """Evaluates the PDFs for the given event data. The particular PDF is
        selected based on the grid parameter values for each model.
        This method is called by the interpolation method.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial event data.
        eventdata : instance of numpy ndarray
            The (V,N_values)-shaped numpy ndarray holding the event data for
            the PDF evaluation.
        gridparams_recarray : instance of numpy structured ndarray
            The numpy structured ndarray of length N_sources with the
            parameter names and values needed for the interpolation on the grid
            for all sources. If the length of this structured array is
            1, the set of parameters will be used for all sources.
        n_values : int
            The size of the output array.

        Returns
        -------
        pd : instance of ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            values for each event.
        """
        logger = get_logger(f'{__name__}.{classname(self)}._evaluate_pdfs')

        # Check for special case when a single set of parameters are provided.
        if len(gridparams_recarray) == 1:
            if self._cfg.is_tracing_enabled:
                logger.debug(
                    'Get PDF for '
                    f'interpol_param_values={gridparams_recarray[0]}.')
            pdf = self._get_pdf_for_interpol_param_values(
                interpol_param_values=gridparams_recarray[0])

            pd = pdf.get_pd_with_eventdata(
                tdm=tdm,
                params_recarray=None,
                eventdata=eventdata,
                tl=tl)

            return pd

        pd = np.empty(n_values, dtype=np.float64)

        (src_idxs, evt_idxs) = tdm.src_evt_idxs

        v_start = 0
        for (sidx, interpol_param_values) in enumerate(gridparams_recarray):
            pdf = self._get_pdf_for_interpol_param_values(
                interpol_param_values=interpol_param_values)

            # Determine the events that belong to the current source.
            evt_mask = src_idxs == sidx

            n = np.count_nonzero(evt_mask)
            sl = slice(v_start, v_start+n)
            pd[sl] = pdf.get_pd_with_eventdata(
                tdm=tdm,
                params_recarray=None,
                eventdata=eventdata,
                evt_mask=evt_mask,
                tl=tl)

            v_start += n

        return pd

    def assert_is_valid_for_trial_data(
            self,
            tdm,
            tl=None,
            **kwargs,
    ):
        """Checks if the PDFs of this PDFSet instance are valid for all the
        given trial data events.
        Since all PDFs should have the same axes, only the first PDF will be
        checked.

        This method calls the
        :meth:`~skyllh.core.pdf.PDFSet.assert_is_valid_for_trial_data` method of
        the :class:`~skyllh.core.pdf.PDFSet` class.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the trial data events.
        tl : instance of TimeLord | None
            The optional instance of TimeLord for measuring timing information.

        Raises
        ------
        ValueError
            If some of the data is outside the axes range of the PDF.
        """
        super().assert_is_valid_for_trial_data(
            tdm=tdm,
            tl=tl,
            **kwargs)

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs,
    ):
        """This method is called whenever a new trial data is initialized. It
        calls the :meth:`~skyllh.core.pdf.PDFSet.initialize_for_new_trial`
        method. Furthermore, it creates the two-dimensional (V,N_events)-shaped
        ndarray holding the event data and stores it to
        ``self._cache_eventdata``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the new trial data events.
        tl : instance of TimeLord | None
            The optional instance of TimeLord for measuring timing information.
        """
        super().initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)

        # Get the first PDF of this PDFSet.
        pdf = next(iter(self.items()))[1]

        with TaskTimer(tl, 'Create MultiDimGridPDFSet eventdata.'):
            self._cache_eventdata =\
                MultiDimGridPDF.create_eventdata_for_sigpdf(
                    tdm=tdm,
                    axes=pdf.axes)

        if self.use_same_photospline_bfi_for_all_pdfs and\
           self.uses_photospline_SplineTable:
            with TaskTimer(
                    tl,
                    'Get and set basis function indices for all PDFs.'):
                V = self._cache_eventdata.shape[0]
                try:
                    bfi = pdf.pdf.search_centers(
                        [self._cache_eventdata[i] for i in range(0, V)]
                    )
                    for (_, pdf) in self.items():
                        pdf.basis_function_indices = bfi
                except ValueError:
                    logger = get_logger(f'{__name__}.{classname(self)}.initialize_for_new_trial')
                    logger.info("Falling back to the slower photospline evaluation.")

    def get_pd(
            self,
            tdm,
            params_recarray,
            tl=None,
    ):
        """Calculates the probability density for each event, given the given
        parameter values.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that will be used to get the data
            from the trial events.
        params_recarray : instance of structured ndarray | None
            The numpy record ndarray holding the parameter name and values for
            each source model.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to use for measuring timing
            information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            value for each source and event.
        grads : dict
            The dictionary holding the PDF gradient value for each event w.r.t.
            each global fit parameter.
            The key of the dictionary is the ID of the global fit parameter.
            The value is the (N_values,)-shaped numpy ndarray holding the
            gradient value for each event.
        """
        logger = get_logger(f'{__name__}.{classname(self)}.get_pd')

        # Get the interpolated PDF values for the arbitrary parameter values.
        # The (D,N_events)-shaped grads_arr ndarray contains the gradient of the
        # probability density w.r.t. each of the D parameters, which are defined
        # by the param_grid_set. The order of the D gradients is the same as
        # the parameter grids.
        with TaskTimer(
                tl,
                'Call interpolate method to get probability densities for all '
                'events.'):
            if self._cfg.is_tracing_enabled:
                logger.debug(
                    'Call interpol_method with '
                    f'params_recarray={params_recarray} of fields '
                    f'{list(params_recarray.dtype.fields.keys())}.')
            (pd, grads_arr) = self._interpol_method(
                tdm=tdm,
                eventdata=self._cache_eventdata,
                params_recarray=params_recarray,
                tl=tl,
            )

        # Construct the gradients dictionary with all the fit parameters, that
        # contribute to the local interpolation parameters.
        grads = dict()

        tdm_n_sources = tdm.n_sources
        for fitparam_id in range(self.pmm.n_global_floating_params):
            grad = np.zeros((tdm.get_n_values(),), dtype=np.float64)

            # Loop through the local interpolation parameters and match them
            # with the global fit parameter fitparam_id.
            fitparam_id_contributes = False
            for (pidx, pname) in enumerate(self._interpol_param_names):
                if pname not in params_recarray.dtype.fields:
                    continue
                p_gpidxs = params_recarray[f'{pname}:gpidx']
                src_mask = p_gpidxs == (fitparam_id + 1)
                n_sources = np.count_nonzero(src_mask)
                if n_sources == 0:
                    continue

                fitparam_id_contributes = True

                if n_sources == tdm_n_sources:
                    # This parameter applies to all sources, hence to all
                    # values, and hence it's the only local parameter
                    # contributing to the global parameter fitparam_id.
                    grad = grads_arr[pidx]
                    break

                # The current parameter does not apply to all sources.
                # Create a values mask that matches a given source mask.
                values_mask = tdm.get_values_mask_for_source_mask(src_mask)
                grad[values_mask] = grads_arr[pidx][values_mask]

            if fitparam_id_contributes:
                grads[fitparam_id] = grad

        return (pd, grads)


class SignalSHGMappedMultiDimGridPDFSet(
        IsSignalPDF,
        PDFSet,
        PDF,
):
    """This class provides a set of MultiDimGridPDF instances, one for one or
    more source hypothesis groups.
    """

    def __init__(
            self,
            shg_mgr,
            pmm,
            shgidxs_pdf_list,
            use_same_photospline_bfi_for_all_pdfs=False,
            **kwargs,
    ):
        """Creates a new SignalSHGMappedMultiDimGridPDFSet instance, which holds
        a set of MultiDimGridPDF instances, one for one or more source
        hypothesis groups.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypothesis groups and their sources.
        pmm : instance of ParameterModelMapper
            The instance of ParameterModelMapper which defines the mapping of
            global parameters to local source parameters.
        shgidxs_pdf_list : sequence of (shg_idxs, MultiDimGridPDF) tuples
            The sequence of 2-element tuples which define the mapping of the
            source hypothesis groups to a PDF instance.
        use_same_photospline_bfi_for_all_pdfs : bool
            Flag if the same basis function indices (bfi) should be used for
            all PDFs when photospline tables are used. Default is ``False``.
            This should be set to ``True`` if all photospline tables share the
            same binning. In that case the PDF evaluation will be accelerated.
        """
        super().__init__(
            pmm=pmm,
            param_set=None,
            param_grid_set=None,
            **kwargs)

        if not isinstance(shg_mgr, SourceHypoGroupManager):
            raise TypeError(
                'The shg_mgr argument must be an instance of '
                'SourceHypoGroupManager! '
                f'Its current type is {classname(shg_mgr)}.')
        self._shg_mgr = shg_mgr

        self._shgidxs_list = []
        for (shg_idxs, pdf) in shgidxs_pdf_list:
            self._shgidxs_list.append(shg_idxs)
            self.add_pdf(
                pdf=pdf,
                gridparams={'shg_idxs': shg_idxs})

        self.use_same_photospline_bfi_for_all_pdfs =\
            use_same_photospline_bfi_for_all_pdfs

        self._cache_eventdata = None

        # Determine if the PDFs are internally represented as
        # photospline.SplineTable instances.
        self.uses_photospline_SplineTable = False
        pdf = next(iter(self.items()))[1]
        if tool.is_available('photospline') and\
           isinstance(pdf.pdf, tool.get('photospline').SplineTable):
            self.uses_photospline_SplineTable = True

    @property
    def shg_mgr(self):
        """(read-only) The instance of SourceHypoGroupManager that defines the
        source hypothesis groups and their sources.
        """
        return self._shg_mgr

    def initialize_for_new_trial(
            self,
            tdm,
            tl=None,
            **kwargs,
    ):
        """This method is called whenever a new trial data is initialized. It
        calls the :meth:`~skyllh.core.pdf.PDFSet.initialize_for_new_trial`
        method. Furthermore, it creates the two-dimensional (V,N_events)-shaped
        ndarray holding the event data and stores it to
        ``self._cache_eventdata``.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager holding the new trial data events.
        tl : instance of TimeLord | None
            The optional instance of TimeLord for measuring timing information.
        """
        super().initialize_for_new_trial(
            tdm=tdm,
            tl=tl,
            **kwargs)

        # Get the first PDF of this PDFSet.
        pdf = next(iter(self.items()))[1]

        with TaskTimer(tl, 'Create MultiDimGridPDFSet eventdata.'):
            # All PDFs of this PDFSet should have the same axes, so we use
            # the axes from the first PDF in this PDF set.
            self._cache_eventdata =\
                MultiDimGridPDF.create_eventdata_for_sigpdf(
                    tdm=tdm,
                    axes=pdf.axes)

        if self.use_same_photospline_bfi_for_all_pdfs and\
           self.uses_photospline_SplineTable:
            with TaskTimer(
                    tl,
                    'Get and set basis function indices for all PDFs.'):
                V = self._cache_eventdata.shape[0]
                try:
                    bfi = pdf.pdf.search_centers(
                        [self._cache_eventdata[i] for i in range(0, V)]
                    )
                    for (_, pdf) in self.items():
                        pdf.basis_function_indices = bfi
                except ValueError:
                    logger = get_logger(f'{__name__}.{classname(self)}.initialize_for_new_trial')
                    logger.info("Falling back to the slower photospline evaluation.")

    def get_pd(
            self,
            tdm,
            params_recarray,
            tl=None,
    ):
        """Calculates the probability density for each event, given the given
        parameter values.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that will be used to get the data
            from the trial events.
        params_recarray : instance of structured ndarray | None
            The numpy record ndarray holding the parameter name and values for
            each source model.
        tl : instance of TimeLord | None
            The optional instance of TimeLord to use for measuring timing
            information.

        Returns
        -------
        pd : instance of numpy ndarray
            The (N_values,)-shaped numpy ndarray holding the probability density
            value for each event.
        grads : dict
            The dictionary holding the PDF gradient value for each event w.r.t.
            each global fit parameter.
            The key of the dictionary is the ID of the global fit parameter.
            The value is the (N_values,)-shaped numpy ndarray holding the
            gradient value for each event.
            By definition this PDF set does not depend on any fit parameters,
            hence, this dictionary is empty.
        """
        pd = np.zeros((tdm.get_n_values(),), dtype=np.float64)

        src_idxs = tdm.src_evt_idxs[0]
        src_idxs_arr = np.arange(self._shg_mgr.n_sources)

        # Loop over the individual PDFs (via their key).
        for shg_idxs in self._shgidxs_list:

            src_mask = np.zeros((self._shg_mgr.n_sources,), dtype=np.bool_)
            for shg_idx in shg_idxs:
                src_mask |= self._shg_mgr.get_src_mask_of_shg(shg_idx)
            pdf_src_idxs = src_idxs_arr[src_mask]
            values_mask = np.isin(src_idxs, pdf_src_idxs)

            pdf_key = self.make_key({'shg_idxs': shg_idxs})
            pdf = self.get_pdf(pdf_key)

            with TaskTimer(tl, f'Get PD values for PDF of SHG {shg_idxs}.'):
                pd_pdf = pdf.get_pd_with_eventdata(
                    tdm=tdm,
                    params_recarray=params_recarray,
                    eventdata=self._cache_eventdata,
                    evt_mask=values_mask)

            pd[values_mask] = pd_pdf

        return (pd, dict())



# -*- coding: utf-8 -*-

"""Plotting module to plot IceCube specific PDF objects.
"""

import itertools

import numpy as np

from matplotlib.axes import (
    Axes,
)
from matplotlib.colors import (
    LogNorm,
)

from skyllh.core.py import (
    classname,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.storage import (
    DataFieldRecordArray,
)
from skyllh.core.trialdata import (
    TrialDataManager,
)
from skyllh.i3.pdf import (
    I3EnergyPDF,
)


class I3EnergyPDFPlotter(object):
    """Plotter class to plot an I3EnergyPDF object.
    """
    def __init__(self, tdm, pdf):
        """Creates a new plotter object for plotting an I3EnergyPDF object.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The instance of TrialDataManager that provides the data for the
            PDF evaluation.
        pdf : I3EnergyPDF
            The PDF object to plot.
        """
        self.tdm = tdm
        self.pdf = pdf

    @property
    def pdf(self):
        """The PDF object to plot.
        """
        return self._pdf

    @pdf.setter
    def pdf(self, obj):
        if not isinstance(obj, I3EnergyPDF):
            raise TypeError(
                'The pdf property must be an object of instance I3EnergyPDF!')
        self._pdf = obj

    @property
    def tdm(self):
        """The TrialDataManager that provides the data for the PDF evaluation.
        """
        return self._tdm

    @tdm.setter
    def tdm(self, obj):
        if not isinstance(obj, TrialDataManager):
            raise TypeError(
                'The tdm property must be an instance of TrialDataManager!')
        self._tdm = obj

    def plot(self, src_hypo_group_manager, axes, **kwargs):
        """Plots the PDF object.

        Parameters
        ----------
        src_hypo_group_manager : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the source
            hypotheses.
        axes : mpl.axes.Axes
            The matplotlib Axes object on which the PDF ratio should get drawn
            to.
        fitparams : dict
            The dictionary with the set of fit paramater values.

        Additional Keyword Arguments
        ----------------------------
        Any additional keyword arguments will be passed to the `mpl.imshow`
        function.

        Returns
        -------
        img : instance of mpl.AxesImage
            The AxesImage instance showing the PDF ratio image.
        """
        if not isinstance(src_hypo_group_manager, SourceHypoGroupManager):
            raise TypeError(
                'The src_hypo_group_manager argument must be an instance of '
                'SourceHypoGroupManager!')
        if not isinstance(axes, Axes):
            raise TypeError(
                'The axes argument must be an instance of '
                'matplotlib.axes.Axes!')

        # The I3EnergyPDF object has two axes, one for log10_energy and sin_dec.
        (xbinning, ybinning) = self._pdf.binnings

        pdf_values = np.zeros((xbinning.nbins, ybinning.nbins), dtype=np.float64)
        events = DataFieldRecordArray(np.zeros(
            (pdf_values.size,),
            dtype=[('ix', np.int64), (xbinning.name, np.float64),
                   ('iy', np.int64), (ybinning.name, np.float64)]))
        for (i, ((ix, x), (iy, y))) in enumerate(itertools.product(
                enumerate(xbinning.bincenters),
                enumerate(ybinning.bincenters))):
            events['ix'][i] = ix
            events[xbinning.name][i] = x
            events['iy'][i] = iy
            events[ybinning.name][i] = y

        self._tdm.initialize_for_new_trial(src_hypo_group_manager, events)

        event_pdf_values = self._pdf.get_prob(self._tdm)
        pdf_values[events['ix'], events['iy']] = event_pdf_values

        (left, right, bottom, top) = (xbinning.lower_edge, xbinning.upper_edge,
                                      ybinning.lower_edge, ybinning.upper_edge)
        img = axes.imshow(
            pdf_values.T,
            extent=(left, right, bottom, top),
            origin='lower',
            norm=LogNorm(),
            interpolation='none',
            **kwargs)
        axes.set_xlabel(xbinning.name)
        axes.set_ylabel(ybinning.name)
        axes.set_title(classname(self._pdf))

        return img



# -*- coding: utf-8 -*-

import argparse
import socket
import time

from skyllh.cluster.commands import (
    ACK,
    MSG,
    RegisterCN,
    ShutdownCN,
    receive_command_from_socket,
)
from skyllh.core.py import (
    int_cast,
)


class ComputeNode(object):
    """The ComputeNode class provides an entity for stand-alone program running
    on a dedicated compute node host.
    """
    def __init__(self, live_time, master_addr, master_port):
        super(ComputeNode, self).__init__()

        self.live_time = live_time
        self.master_addr = master_addr
        self.master_port = master_port

        self._start_time = time.time()

        # Register the compute node to the master node.
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.master_addr, self.master_port))

        # Send the register command to the master and tell .
        RegisterCN(self._start_time, self._live_time).send(self.sock)
        reply = receive_command_from_socket(self.sock)
        if not reply.is_same_as(ACK):
            raise RuntimeError(
                'The master node did not reply with an ACK command!')

        print(f'Registered to master {self._master_addr}:{self._master_port}')
        print(f'Runtime set to {self._live_time} seconds')

    def __del__(self):
        self.sock.close()

    @property
    def live_time(self):
        """The time in seconds this ComputeNode instance should be listening for
        requests.
        """
        return self._live_time

    @live_time.setter
    def live_time(self, t):
        t = int_cast(
            t,
            'The live_time property must be castable to type int!')
        self._live_time = t

    @property
    def master_addr(self):
        """The address of the SkyLLH master program.
        """
        return self._master_addr

    @master_addr.setter
    def master_addr(self, addr):
        if not isinstance(addr, str):
            raise TypeError(
                'The master_addr property must be of type str!')
        self._master_addr = addr

    @property
    def master_port(self):
        """The port number of the SkyLLH master program.
        """
        return self._master_port

    @master_port.setter
    def master_port(self, p):
        p = int_cast(
            p,
            'The master_port property must be castable to type int!')
        self._master_port = p

    def handle_requests(self):
        if time.time() > self._start_time + self._live_time:
            raise RuntimeError('Live-time already exceeded!')

        while True:
            # Receive a command.
            cmd = receive_command_from_socket(self.sock)
            if cmd.is_same_as(MSG):
                print(f'Received general message: {cmd.msg}')
            elif cmd.is_same_as(ShutdownCN):
                print('Received shutdown command. Shutting down.')
                self.sock.close()
                return
            else:
                print('Received unknown command! Ignoring.')

            if time.time() > self._start_time + self._live_time:
                print('Live-time exceeded. Shutting down.')
                return


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='SkyLLH Compute Node')
    parser.add_argument(
        'master_addr', type=str,
        help='The address (IP / hostname) of the SkyLLH master program.')
    parser.add_argument(
        'master_port', type=int, default=9999,
        help='The port number of the SkyLLH master program.')
    parser.add_argument(
        '--live-time', type=int, default=2*60*60,
        help='The time in seconds to run this compute node instance.')

    args = parser.parse_args()

    cn = ComputeNode(
        live_time=args.live_time,
        master_addr=args.master_addr,
        master_port=args.master_port)

    cn.handle_requests()



# -*- coding: utf-8 -*-

import abc
import inspect
import numpy as np
import scipy.sparse

from skyllh.core.py import (
    classname,
    float_cast,
    issequenceof,
)
from skyllh.core.source_hypo_grouping import (
    SourceHypoGroupManager,
)
from skyllh.core.source_model import (
    SourceModel,
)
from skyllh.core.timing import (
    TaskTimer,
)
from skyllh.core.utils.coords import (
    angular_separation,
)


class EventSelectionMethod(
        object,
        metaclass=abc.ABCMeta):
    """This is the abstract base class for all event selection method classes.
    The idea is to pre-select only events that contribute to the likelihood
    function, i.e. are more signal than background like. The different methods
    are implemented through derived classes of this base class.
    """

    def __init__(
            self,
            shg_mgr,
            **kwargs):
        """Creates a new event selection method instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager | None
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
            It can be ``None`` if the event selection method does not depend on
            the sources.
        """
        super().__init__(
            **kwargs)

        self._src_arr = None

        self._shg_mgr = shg_mgr
        if self._shg_mgr is not None:
            if not isinstance(self._shg_mgr, SourceHypoGroupManager):
                raise TypeError(
                    'The shg_mgr argument must be None or an instance of '
                    'SourceHypoGroupManager! '
                    f'Its current type is {classname(self._shg_mgr)}.')

            # The _src_arr variable holds a numpy record array with the
            # necessary source information needed for the event selection
            # method.
            self._src_arr = self.sources_to_array(
                sources=self._shg_mgr.source_list)

    @property
    def shg_mgr(self):
        """(read-only) The instance of SourceHypoGroupManager, which defines the
        list of sources.
        """
        return self._shg_mgr

    def __and__(self, other):
        """Implements the AND operator (&) for creating an event selection
        method, which is the intersection of this event selection method and
        another one using the expression ``intersection = self & other``.

        Parameters
        ----------
        other : instance of EventSelectionMethod
            The instance of EventSelectionMethod that is the other event
            selection method.

        Returns
        -------
        intersection : instance of IntersectionEventSelectionMethod
            The instance of IntersectionEventSelectionMethod that creates the
            intersection of this event selection method and the other.
        """
        return IntersectionEventSelectionMethod(self, other)

    def change_shg_mgr(self, shg_mgr):
        """Changes the SourceHypoGroupManager instance of the event selection
        method. This will also recreate the internal source numpy record array.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager | None
            The new SourceHypoGroupManager instance, that should be used for
            this event selection method.
            It can be ``None`` if the event selection method does not depend on
            the sources.
        """
        self._shg_mgr = shg_mgr
        self._src_arr = None

        if self._shg_mgr is not None:
            if not isinstance(self._shg_mgr, SourceHypoGroupManager):
                raise TypeError(
                    'The shg_mgr argument must be None or an instance of '
                    'SourceHypoGroupManager! '
                    f'Its current type is {classname(self._shg_mgr)}.')

            self._src_arr = self.sources_to_array(
                sources=self._shg_mgr.source_list)

    def sources_to_array(self, sources):
        """This method is supposed to convert a sequence of SourceModel
        instances into a structured numpy ndarray with the source information
        in a format that is best understood by the actual event selection
        method.

        Parameters
        ----------
        sources : sequence of SourceModel
            The sequence of source models containing the necessary information
            of the source.

        Returns
        -------
        arr : numpy record ndarray | None
            The generated numpy record ndarray holding the necessary information
            for each source.
            By default ``None`` is returned.
        """
        return None

    @abc.abstractmethod
    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """This method selects the events, which will contribute to the
        log-likelihood ratio function.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray of length N_events, holding the
            events.
        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray of length N_selected_events,
            holding the selected events, i.e. a subset of the ``events``
            argument.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The two 1d ndarrays of int of length N_values, holding the indices
            of the sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        pass


class IntersectionEventSelectionMethod(
        EventSelectionMethod):
    """This class provides an event selection method for the intersection of two
    event selection methods. It can be created using the ``&`` operator:
    ``evt_sel_method1 & evt_sel_method2``.
    """
    def __init__(
            self,
            evt_sel_method1,
            evt_sel_method2,
            **kwargs):
        """Creates a compounded event selection method of two given event
        selection methods.

        Parameters
        ----------
        evt_sel_method1 : instance of EventSelectionMethod
            The instance of EventSelectionMethod for the first event selection
            method.
        evt_sel_method2 : instance of EventSelectionMethod
            The instance of EventSelectionMethod for the second event selection
            method.
        """
        super().__init__(
            shg_mgr=None,
            **kwargs)

        self.evt_sel_method1 = evt_sel_method1
        self.evt_sel_method2 = evt_sel_method2

    @property
    def evt_sel_method1(self):
        """The instance of EventSelectionMethod for the first event selection
        method.
        """
        return self._evt_sel_method1

    @evt_sel_method1.setter
    def evt_sel_method1(self, method):
        if not isinstance(method, EventSelectionMethod):
            raise TypeError(
                'The evt_sel_method1 property must be an instance of '
                'EventSelectionMethod!'
                f'Its current type is {classname(method)}.')
        self._evt_sel_method1 = method

    @property
    def evt_sel_method2(self):
        """The instance of EventSelectionMethod for the second event selection
        method.
        """
        return self._evt_sel_method2

    @evt_sel_method2.setter
    def evt_sel_method2(self, method):
        if not isinstance(method, EventSelectionMethod):
            raise TypeError(
                'The evt_sel_method2 property must be an instance of '
                'EventSelectionMethod!'
                f'Its current type is {classname(method)}.')
        self._evt_sel_method2 = method

    def change_shg_mgr(self, shg_mgr):
        """Changes the SourceHypoGroupManager instance of the event selection
        method. This will call the ``change_shg_mgr`` of the individual event
        selection methods.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager | None
            The new SourceHypoGroupManager instance, that should be used for
            this event selection method.
            It can be ``None`` if the event selection method does not depend on
            the sources.
        """
        self._evt_sel_method1.change_shg_mgr(shg_mgr=shg_mgr)
        self._evt_sel_method2.change_shg_mgr(shg_mgr=shg_mgr)

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects events by calling the ``select_events`` methods of the
        individual event selection methods.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the events.
        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : DataFieldRecordArray
            The instance of DataFieldRecordArray holding the selected events,
            i.e. a subset of the `events` argument.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The indices of the sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        if ret_original_evt_idxs:
            (events, src_evt_idxs, org_evt_idxs1) =\
                self._evt_sel_method1.select_events(
                    events=events,
                    src_evt_idxs=src_evt_idxs,
                    ret_original_evt_idxs=True)

            (events, src_evt_idxs, org_evt_idxs2) =\
                self._evt_sel_method2.select_events(
                    events=events,
                    src_evt_idxs=src_evt_idxs,
                    ret_original_evt_idxs=True)

            org_evt_idxs = np.take(org_evt_idxs1, org_evt_idxs2)

            return (events, src_evt_idxs, org_evt_idxs)

        (events, src_evt_idxs) = self._evt_sel_method1.select_events(
            events=events,
            src_evt_idxs=src_evt_idxs)

        (events, src_evt_idxs) = self._evt_sel_method2.select_events(
            events=events,
            src_evt_idxs=src_evt_idxs)

        return (events, src_evt_idxs)


class AllEventSelectionMethod(
        EventSelectionMethod):
    """This event selection method selects all events.
    """
    def __init__(self, shg_mgr):
        """Creates a new event selection method instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances. For this particular
            event selection method it has no meaning, but it is an interface
            parameter.
        """
        super().__init__(
            shg_mgr=shg_mgr)

    def sources_to_array(self, sources):
        """Creates the source array from the given list of sources. This event
        selection method does not depend on the sources. Hence, ``None`` is
        returned.

        Returns
        -------
        arr : None
            The generated numpy record ndarray holding the necessary information
            for each source. Since this event selection method does not depend
            on any source, ``None`` is returned.
        """
        return None

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects all of the given events. Hence, the returned event array is
        the same as the given array.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the events, for which
            the selection method should get applied.
        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : DataFieldRecordArray
            The instance of DataFieldRecordArray holding the selected events,
            i.e. a subset of the `events` argument.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The indices of sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        with TaskTimer(tl, 'ESM: Calculate indices of selected events.'):
            if src_evt_idxs is None:
                n_sources = self.shg_mgr.n_sources
                src_idxs = np.repeat(np.arange(n_sources), len(events))
                evt_idxs = np.tile(events.indices, n_sources)
            else:
                (src_idxs, evt_idxs) = src_evt_idxs

        if ret_original_evt_idxs:
            return (events, (src_idxs, evt_idxs), events.indices)

        return (events, (src_idxs, evt_idxs))


class SpatialEventSelectionMethod(
        EventSelectionMethod,
        metaclass=abc.ABCMeta):
    """This abstract base class defines the base class for all spatial event
    selection methods.
    """

    def __init__(
            self,
            shg_mgr,
            **kwargs):
        """Creates a new event selection method instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            **kwargs)

    def sources_to_array(self, sources):
        """Converts the given sequence of SourceModel instances into a
        structured numpy ndarray holding the necessary source information needed
        for this event selection method.

        Parameters
        ----------
        sources : sequence of SourceModel
            The sequence of source models containing the necessary information
            of the source.

        Returns
        -------
        arr : numpy record ndarray
            The generated numpy record ndarray holding the necessary information
            for each source. It contains the following data fields: 'ra', 'dec'.
        """
        if not issequenceof(sources, SourceModel):
            raise TypeError(
                'The sources argument must be a sequence of SourceModel '
                'instances! '
                f'Its current type is {classname(sources)}.')

        arr = np.empty(
            (len(sources),),
            dtype=[
                ('ra', np.float64),
                ('dec', np.float64)
            ],
            order='F')

        for (i, src) in enumerate(sources):
            arr['ra'][i] = src.ra
            arr['dec'][i] = src.dec

        return arr


class DecBandEventSectionMethod(
        SpatialEventSelectionMethod):
    """This event selection method selects events within a declination band
    around a list of point-like source positions.
    """
    def __init__(
            self,
            shg_mgr,
            delta_angle):
        """Creates and configures a spatial declination band event selection
        method object.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source in declination for which
            events should get selected.
        """
        super().__init__(
            shg_mgr=shg_mgr)

        self.delta_angle = delta_angle

    @property
    def delta_angle(self):
        """The half-opening angle around the source in declination and
        right-ascention for which events should get selected.
        """
        return self._delta_angle

    @delta_angle.setter
    def delta_angle(self, angle):
        angle = float_cast(
            angle,
            'The delta_angle property must be castable to type float!')
        self._delta_angle = angle

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects the events within the declination band.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

                ``'dec'`` : float
                    The declination of the event.

        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The indices of sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr

        # Calculates the minus and plus declination around each source and
        # bound it to -90deg and +90deg, respectively.
        src_dec_minus = np.maximum(-np.pi/2, src_arr['dec'] - delta_angle)
        src_dec_plus = np.minimum(src_arr['dec'] + delta_angle, np.pi/2)

        # Determine the mask for the events which fall inside the declination
        # window.
        # mask_dec is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM-DecBand: Calculate mask_dec.'):
            mask_dec = (
                (events['dec'] > src_dec_minus[:, np.newaxis]) &
                (events['dec'] < src_dec_plus[:, np.newaxis])
            )

        # Determine the mask for the events that fall inside at least one
        # source declination band.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM-DecBand: Calculate mask.'):
            mask = np.any(mask_dec, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM-DecBand: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events_idxs = events.indices[mask]
            selected_events = events[selected_events_idxs]

        # Get selected events indices.
        idxs = np.argwhere(mask_dec[:, mask])
        src_idxs = idxs[:, 0]
        evt_idxs = idxs[:, 1]

        if ret_original_evt_idxs:
            return (selected_events, (src_idxs, evt_idxs), selected_events_idxs)

        return (selected_events, (src_idxs, evt_idxs))


class RABandEventSectionMethod(
        SpatialEventSelectionMethod):
    """This event selection method selects events within a right-ascension band
    around a list of point-like source positions.
    """
    def __init__(
            self,
            shg_mgr,
            delta_angle):
        """Creates and configures a right-ascension band event selection
        method object.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source in right-ascension for
            which events should get selected.
        """
        super().__init__(
            shg_mgr=shg_mgr)

        self.delta_angle = delta_angle

    @property
    def delta_angle(self):
        """The half-opening angle around the source in declination and
        right-ascention for which events should get selected.
        """
        return self._delta_angle

    @delta_angle.setter
    def delta_angle(self, angle):
        angle = float_cast(
            angle,
            'The delta_angle property must be castable to type float!')
        self._delta_angle = angle

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects the events within the right-ascention band.

        The solid angle dOmega = dRA * dSinDec = dRA * dDec * cos(dec) is a
        function of declination, i.e. for a constant dOmega, the right-ascension
        value has to change with declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            ``'ra'`` : float
                The right-ascention of the event.
            ``'dec'`` : float
                The declination of the event.

        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The indices of the sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr

        # Get the minus and plus declination around the sources.
        src_dec_minus = np.maximum(-np.pi/2, src_arr['dec'] - delta_angle)
        src_dec_plus = np.minimum(src_arr['dec'] + delta_angle, np.pi/2)

        # Calculate the cosine factor for the largest declination distance from
        # the source. We use np.amin here because smaller cosine values are
        # larger angles.
        # cosfact is a (N_sources,)-shaped ndarray.
        cosfact = np.amin(np.cos([src_dec_minus, src_dec_plus]), axis=0)

        # Calculate delta RA, which is a function of declination.
        # dRA is a (N_sources,)-shaped ndarray.
        dRA_half = np.amin(
            [np.repeat(2*np.pi, len(src_arr['ra'])),
             np.fabs(delta_angle / cosfact)], axis=0)

        # Calculate the right-ascension distance of the events w.r.t. the
        # source. We make sure to use the smaller distance on the circle, thus
        # the maximal distance is 180deg, i.e. pi.
        # ra_dist is a (N_sources,N_events)-shaped 2D ndarray.
        with TaskTimer(tl, 'ESM-RaBand: Calculate ra_dist.'):
            ra_dist = np.fabs(
                np.mod(
                    events['ra'] - src_arr['ra'][:, np.newaxis] + np.pi,
                    2*np.pi) - np.pi)

        # Determine the mask for the events which fall inside the
        # right-ascention window.
        # mask_ra is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM-RaBand: Calculate mask_ra.'):
            mask_ra = ra_dist < dRA_half[:, np.newaxis]

        # Determine the mask for the events that fall inside at least one
        # source sky window.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM-RaBand: Calculate mask.'):
            mask = np.any(mask_ra, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM-RaBand: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events_idxs = events.indices[mask]
            selected_events = events[selected_events_idxs]

        # Get selected events indices.
        idxs = np.argwhere(mask_ra[:, mask])
        src_idxs = idxs[:, 0]
        evt_idxs = idxs[:, 1]

        if ret_original_evt_idxs:
            return (selected_events, (src_idxs, evt_idxs), selected_events_idxs)

        return (selected_events, (src_idxs, evt_idxs))


class SpatialBoxEventSelectionMethod(
        SpatialEventSelectionMethod):
    """This event selection method selects events within a spatial box in
    right-ascention and declination around a list of point-like source
    positions.
    """
    def __init__(
            self,
            shg_mgr,
            delta_angle):
        """Creates and configures a spatial box event selection method object.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source for which events should
            get selected.
        """
        super().__init__(
            shg_mgr=shg_mgr)

        self.delta_angle = delta_angle

    @property
    def delta_angle(self):
        """The half-opening angle around the source in declination and
        right-ascention for which events should get selected.
        """
        return self._delta_angle

    @delta_angle.setter
    def delta_angle(self, angle):
        angle = float_cast(
            angle,
            'The delta_angle property must be castable to type float!')
        self._delta_angle = angle

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects the events within the spatial box in right-ascention and
        declination.

        The solid angle dOmega = dRA * dSinDec = dRA * dDec * cos(dec) is a
        function of declination, i.e. for a constant dOmega, the right-ascension
        value has to change with declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            ``'ra'`` : float
                The right-ascention of the event.
            ``'dec'`` : float
                The declination of the event.

        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, evt_idxs) : 1d ndarrays of ints | None
            The indices of sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        delta_angle = self._delta_angle
        src_arr = self._src_arr
        n_sources = len(src_arr)

        srcs_ra = src_arr['ra']
        srcs_dec = src_arr['dec']

        # Get the minus and plus declination around the sources.
        src_dec_minus = np.maximum(-np.pi/2, srcs_dec - delta_angle)
        src_dec_plus = np.minimum(srcs_dec + delta_angle, np.pi/2)

        # Calculate the cosine factor for the largest declination distance from
        # the source. We use np.amin here because smaller cosine values are
        # larger angles.
        # cosfact is a (N_sources,)-shaped ndarray.
        cosfact = np.amin(np.cos([src_dec_minus, src_dec_plus]), axis=0)

        # Calculate delta RA, which is a function of declination.
        # dRA is a (N_sources,)-shaped ndarray.
        dRA_half = np.amin(
            [np.repeat(2*np.pi, n_sources),
             np.fabs(delta_angle / cosfact)], axis=0)

        # Determine the mask for the events which fall inside the
        # right-ascention window.
        # mask_ra is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_ra.'):
            evts_ra = events['ra']
            # Fill in batch sizes of 128 maximum to save memory.
            batch_size = 128
            if n_sources > batch_size:
                mask_ra = np.zeros((n_sources, len(evts_ra)), dtype=bool)
                n_batches = int(np.ceil(n_sources / float(batch_size)))
                for bi in range(n_batches):
                    if bi == n_batches-1:
                        # We got the last batch of sources.
                        srcs_slice = slice(bi*batch_size, None)
                    else:
                        srcs_slice = slice(bi*batch_size, (bi+1)*batch_size)

                    ra_diff = np.fabs(
                        evts_ra - srcs_ra[srcs_slice][:, np.newaxis])
                    ra_mod = np.where(
                        ra_diff >= np.pi, 2*np.pi - ra_diff, ra_diff)
                    mask_ra[srcs_slice, :] = (
                        ra_mod < dRA_half[srcs_slice][:, np.newaxis]
                    )
            else:
                ra_diff = np.fabs(evts_ra - srcs_ra[:, np.newaxis])
                ra_mod = np.where(ra_diff >= np.pi, 2*np.pi-ra_diff, ra_diff)
                mask_ra = ra_mod < dRA_half[:, np.newaxis]

        # Determine the mask for the events which fall inside the declination
        # window.
        # mask_dec is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_dec.'):
            mask_dec = (
                (events['dec'] > src_dec_minus[:, np.newaxis]) &
                (events['dec'] < src_dec_plus[:, np.newaxis])
            )

        # Determine the mask for the events which fall inside the
        # right-ascension and declination window.
        # mask_sky is a (N_sources,N_events)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask_sky.'):
            mask_sky = mask_ra & mask_dec
            del mask_ra
            del mask_dec

        # Determine the mask for the events that fall inside at least one
        # source sky window.
        # mask is a (N_events,)-shaped ndarray.
        with TaskTimer(tl, 'ESM: Calculate mask.'):
            mask = np.any(mask_sky, axis=0)

        # Reduce the events according to the mask.
        with TaskTimer(tl, 'ESM: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events_idxs = events.indices[mask]
            selected_events = events[selected_events_idxs]

        # Get selected events indices.
        idxs = np.argwhere(mask_sky[:, mask])
        src_idxs = idxs[:, 0]
        evt_idxs = idxs[:, 1]

        if ret_original_evt_idxs:
            return (selected_events, (src_idxs, evt_idxs), selected_events_idxs)

        return (selected_events, (src_idxs, evt_idxs))


class PsiFuncEventSelectionMethod(
        EventSelectionMethod):
    """This event selection method selects events whose psi value, i.e. the
    great circle distance of the event to the source, is smaller than the value
    of the provided function.
    """
    def __init__(
            self,
            shg_mgr,
            psi_name,
            func,
            axis_name_list):
        """Creates a new PsiFuncEventSelectionMethod instance.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        psi_name : str
            The name of the data field that provides the psi value of the event.
        func : callable
            The function that should get evaluated for each event. The call
            signature must be

                ``func(*axis_data)``,

            where ``*axis_data`` is the event data of each required axis. The
            number of axes must match the provided axis names through the
            ``axis_name_list``.
        axis_name_list : list of str
            The list of data field names for each axis of the function ``func``.
            All field names must be valid field names of the trial data's
            DataFieldRecordArray instance.
        """
        super().__init__(
            shg_mgr=shg_mgr)

        self.psi_name = psi_name
        self.func = func
        self.axis_name_list = axis_name_list

        n_func_args = len(inspect.signature(self._func).parameters)
        if n_func_args < len(self._axis_name_list):
            raise TypeError(
                'The func argument must be a callable instance with at least '
                f'{len(self._axis_name_list)} arguments! Its current number '
                f'of arguments is {n_func_args}.')

        n_sources = self.shg_mgr.n_sources
        if n_sources != 1:
            raise ValueError(
                'The `PsiFuncEventSelectionMethod.select_events` currently '
                'supports only a single source. It was called with '
                f'{n_sources} sources.')

    @property
    def psi_name(self):
        """The name of the data field that provides the psi value of the event.
        """
        return self._psi_name

    @psi_name.setter
    def psi_name(self, name):
        if not isinstance(name, str):
            raise TypeError(
                'The psi_name property must be an instance of type str! '
                f'Its current type is {classname(name)}.')
        self._psi_name = name

    @property
    def func(self):
        """The function that should get evaluated for each event. The call
        signature must be ``func(*axis_data)``, where ``*axis_data`` is the
        event data of each required axis. The number of axes must match the
        provided axis names through the ``axis_name_list`` property.
        """
        return self._func

    @func.setter
    def func(self, f):
        if not callable(f):
            raise TypeError(
                'The func property must be a callable instance! '
                f'Its current type is {classname(f)}.')
        self._func = f

    @property
    def axis_name_list(self):
        """The list of data field names for each axis of the function defined
        through the ``func`` property.
        """
        return self._axis_name_list

    @axis_name_list.setter
    def axis_name_list(self, names):
        if not issequenceof(names, str):
            raise TypeError(
                'The axis_name_list property must be a sequence of str '
                'instances! '
                f'Its current type is {classname(names)}.')
        self._axis_name_list = list(names)

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects the events whose psi value is smaller than the value of the
        predefined function.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            <psi_name> : float
                The great circle distance of the event with the source.
            <axis_name(s)> : float
                The name of the axis required for the function ``func`` to be
                evaluated.

        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The indices of the sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        cls_name = classname(self)

        with TaskTimer(tl, f'{cls_name}: Get psi values.'):
            psi = events[self._psi_name]

        with TaskTimer(tl, f'{cls_name}: Get axis data values.'):
            func_args = [events[axis] for axis in self._axis_name_list]

        with TaskTimer(tl, f'{cls_name}: Creating mask.'):
            mask = psi < self._func(*func_args)

        with TaskTimer(tl, f'{cls_name}: Create selected_events.'):
            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events_idxs = events.indices[mask]
            selected_events = events[selected_events_idxs]

        # Get selected events indices.
        idxs = np.argwhere(np.atleast_2d(mask))
        src_idxs = idxs[:, 0]
        evt_idxs = idxs[:, 1]

        if ret_original_evt_idxs:
            return (selected_events, (src_idxs, evt_idxs), selected_events_idxs)

        return (selected_events, (src_idxs, evt_idxs))


class AngErrOfPsiEventSelectionMethod(
        SpatialEventSelectionMethod):
    """This event selection method selects events within a spatial box in
    right-ascention and declination around a list of point-like source
    positions and performs an additional selection of events whose ang_err value
    is larger than the value of the provided function at a given psi value.
    """
    def __init__(
            self,
            shg_mgr,
            func,
            psi_floor=None,
            **kwargs):
        """Creates and configures a spatial box and psi func event selection
        method object.

        Parameters
        ----------
        shg_mgr : instance of SourceHypoGroupManager
            The instance of SourceHypoGroupManager that defines the list of
            sources, i.e. the list of SourceModel instances.
        delta_angle : float
            The half-opening angle around the source for which events should
            get selected.
        psi_name : str | None
            The name of the data field that provides the psi value of the event.
            If set to ``None``, the psi value will be calculated automatically.
        func : callable
            The function that should get evaluated for each event. The call
            signature must be

                ``func(psi)``,

            where ``psi`` is the opening angle between the source and the event.
        psi_floor : float | None
            The psi func event selection is excluded for events having psi value
            below the ``psi_floor``. If None, set it to default 5 degrees.
        """
        super().__init__(
            shg_mgr=shg_mgr,
            **kwargs)

        self.func = func

        if psi_floor is None:
            psi_floor = np.deg2rad(5)
        self.psi_floor = psi_floor

    @property
    def func(self):
        """The function that should get evaluated for each event. The call
        signature must be ``func(*axis_data)``, where ``*axis_data`` is the
        event data of each required axis. The number of axes must match the
        provided axis names through the ``axis_name_list`` property.
        """
        return self._func

    @func.setter
    def func(self, f):
        if not callable(f):
            raise TypeError(
                'The func property must be a callable instance! '
                f'Its current type is {classname(f)}.')
        self._func = f

    @property
    def psi_floor(self):
        """The psi func event selection is excluded for events having psi value
        below the `psi_floor`.
        """
        return self._psi_floor

    @psi_floor.setter
    def psi_floor(self, psi):
        psi = float_cast(
            psi,
            'The psi_floor property must be castable to type float!')
        self._psi_floor = psi

    def select_events(
            self,
            events,
            src_evt_idxs=None,
            ret_original_evt_idxs=False,
            tl=None):
        """Selects the events within the spatial box in right-ascention and
        declination and performs an additional selection of events whose ang_err
        value is larger than the value of the provided function at a given psi
        value.

        The solid angle dOmega = dRA * dSinDec = dRA * dDec * cos(dec) is a
        function of declination, i.e. for a constant dOmega, the right-ascension
        value has to change with declination.

        Parameters
        ----------
        events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray that holds the event data.
            The following data fields must exist:

            ``'ra'`` : float
                The right-ascention of the event.
            ``'dec'`` : float
                The declination of the event.

        src_evt_idxs : 2-tuple of 1d ndarrays of ints | None
            The 2-element tuple holding the two 1d ndarrays of int of length
            N_values, specifying to which sources the given events belong to.
            If set to ``None`` all given events will be considered to for all
            sources.
        ret_original_evt_idxs : bool
            Flag if the original indices of the selected events should get
            returned as well.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        selected_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding only the selected
            events.
        (src_idxs, evt_idxs) : 1d ndarrays of ints
            The indices of the sources and the selected events.
        original_evt_idxs : 1d ndarray of ints
            The (N_selected_events,)-shaped numpy ndarray holding the original
            indices of the selected events, if ``ret_original_evt_idxs`` is set
            to ``True``.
        """
        if src_evt_idxs is None:
            n_sources = len(self._src_arr)
            n_events = len(events)
            src_idxs = np.repeat(np.arange(n_sources), n_events)
            evt_idxs = np.tile(np.arange(n_events), n_sources)
        else:
            (src_idxs, evt_idxs) = src_evt_idxs

        # Perform selection based on psi values.
        with TaskTimer(tl, 'ESM: Calculate psi values.'):
            psi = angular_separation(
                ra1=np.take(self._src_arr['ra'], src_idxs),
                dec1=np.take(self._src_arr['dec'], src_idxs),
                ra2=np.take(events['ra'], evt_idxs),
                dec2=np.take(events['dec'], evt_idxs),
            )

        with TaskTimer(tl, 'ESM: Create mask_psi.'):
            mask_psi = (
                (events['ang_err'][evt_idxs] >= self._func(psi)) |
                (psi < self.psi_floor)
            )

        with TaskTimer(tl, 'ESM: Create selected_events.'):
            # Have to define the shape argument in order to not truncate
            # the mask in case last events are not selected.
            mask_sky = scipy.sparse.csr_matrix(
                (mask_psi, (src_idxs, evt_idxs)),
                shape=(len(self._src_arr), len(events))
            ).toarray()
            mask = np.any(mask_sky, axis=0)

            # Using an integer indices array for data selection is several
            # factors faster than using a boolean array.
            selected_events_idxs = events.indices[mask]
            selected_events = events[selected_events_idxs]

        # Get final selected events indices.
        idxs = np.argwhere(mask_sky[:, mask])
        src_idxs = idxs[:, 0]
        evt_idxs = idxs[:, 1]

        if ret_original_evt_idxs:
            return (selected_events, (src_idxs, evt_idxs), selected_events_idxs)

        return (selected_events, (src_idxs, evt_idxs))
